from typing import Iterable, Union

import numpy as np
import numpy.linalg as la
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import KernelCenterer
from sklearn.utils.validation import check_is_fitted

from cca_zoo.models._iterative._gradkcca import GradKCCA
from cca_zoo.utils import _check_views, _process_parameter


class SCCA_HSIC(GradKCCA):
    """
    References
    ----------
    [1] Uurtio, V., Bhadra, S., Rousu, J. Sparse Non-Linear CCA through Hilbert-Schmidt Independence Criterion. IEEE International Conference on Data Mining (ICDM 2018), to appear

    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        proj: Union[Iterable[float], float] = "l1",
        gamma: Iterable[float] = None,
        coef0: Iterable[float] = None,
        repetitions=5,
        initialization: Union[str, callable] = "random",
        nystrom=False,
        nystrom_components=100,
        c=1,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
            initialization=initialization,
            proj=proj,
            gamma=gamma,
            coef0=coef0,
            kernel="rbf",
            repetitions=repetitions,
            nystrom=nystrom,
            nystrom_components=nystrom_components,
        )
        self.c = c

    def _check_params(self):
        self.proj = _process_parameter("proj", self.proj, None, self.n_views)
        self.kernel = _process_parameter("kernel", self.kernel, "linear", self.n_views)
        self.gamma = _process_parameter("gamma", self.gamma, None, self.n_views)
        self.coef0 = _process_parameter("coef0", self.coef0, 1, self.n_views)
        self.degree = _process_parameter("degree", self.degree, 1, self.n_views)
        self.c = _process_parameter("c", self.c, 1, self.n_views)

    def _objective(self, views, scores, weights) -> int:
        N = scores[0].shape[0]
        return np.trace(scores[0] @ scores[1]).sum() / (N - 1) ** 2

    def backracking_line_search(self, w, gw, stp, X, cKv, obj_old, view_index):
        while True:
            w_new = w + gw * stp
            if self.proj[view_index] == "l1":
                w_new = self._proj_l1(w_new, self.c[view_index])
            elif self.proj[view_index] == "l2":
                w_new = self._proj_l2(w_new, self.c[view_index])
            else:
                raise ValueError(
                    "projection {self.proj[view_index]} not supported. Pass a generator implementing this method"
                )
            Kw = self._get_kernel(view_index, X @ w_new[:, None])
            obj_new = self._objective(None, (Kw, cKv), None)
            if obj_new > obj_old + 1e-4 * np.abs(obj_old):
                return w_new
            elif stp < 1e-7:
                return w
            else:
                stp /= 2

    def _update(self, views, scores, weights):
        flipped_idxs = [1, 0]
        K = [
            self._get_kernel(i, view @ weights[i][:, None])
            for i, view in enumerate(views)
        ]
        cK = [KernelCenterer().fit_transform(K_) for K_ in K]
        weights = [
            self._proj_l2(w, self.c[view_index]) for view_index, w in enumerate(weights)
        ]
        for view_index, (view, flipped_idx) in enumerate(zip(views, flipped_idxs)):
            obj_old = self._objective(views, (K[view_index], cK[flipped_idx]), weights)
            grad = gradf_gauss_sgd(
                K[view_index],
                cK[flipped_idx],
                view,
                self.gamma[view_index],
                weights[view_index],
            )
            gamma = la.norm(grad)
            weights[view_index] = self.backracking_line_search(
                weights[view_index],
                grad,
                gamma,
                view,
                cK[flipped_idx],
                obj_old,
                view_index,
            )
            K[view_index] = self._get_kernel(
                view_index, view @ weights[view_index][:, None]
            )
            cK[view_index] = KernelCenterer().fit_transform(K[view_index])
        return K, weights

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        """

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        check_is_fitted(self, attributes=["weights"])
        views = _check_views(*views)
        views = self._centre_scale_transform(views)
        transformed_views = []
        for i, (view) in enumerate(views):
            transformed_view = view @ self.weights[i]
            transformed_views.append(transformed_view)
        return transformed_views


def gradf_gauss_sgd(K1, cK2, X, a, u):
    if a is None:
        a = 1 / X.shape[1]
    N = K1.shape[0]
    temp = 0
    id1 = np.argsort(np.random.rand(N))[: int(N / 10)]
    id2 = np.argsort(np.random.rand(N))[: int(N / 10)]
    for i in id1:
        for j in id2:
            temp += (
                K1[i, j] * cK2[i, j] * np.outer(X[i, :] - X[j, :], X[i, :] - X[j, :])
            )
    return -(2 * a * u.T @ temp).T


def generate_data(n, p, q):
    X = np.random.uniform(-1, 1, [n, p])
    Y = np.random.uniform(-1, 1, [n, q])
    Y[:, 2] = X[:, 2] + X[:, 3] - Y[:, 3] + np.random.normal(0, 0.05, n)
    # Y[:,2] = np.power(X[:,2] + X[:,3],3) - Y[:,3] + np.random.normal(0,0.05,n)
    # Y[:,4] = np.exp(X[:,4] + X[:,5]) - Y[:,5] + np.random.normal(0,0.05,n)
    return X, Y


def main():
    import matplotlib.pyplot as plt

    np.set_printoptions(precision=2)

    X, Y = generate_data(1000, 8, 8)
    Xtrain, Xtest, Ytrain, Ytest = train_test_split(X, Y)

    kernel1 = SCCA_HSIC(proj="l2").fit((Xtrain, Ytrain))
    print(f"Training Correlation: {kernel1.score((Xtrain, Ytrain))}")
    print(f"Test Correlation: {kernel1.score((Xtest, Ytest))}")

    plt.plot(*kernel1.transform((Xtrain, Ytrain)), "bo")
    plt.show()


if __name__ == "__main__":
    main()



from typing import Iterable

import numpy as np

from cca_zoo.models._base import _BaseCCA


class _DummyCCA(_BaseCCA):
    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        accept_sparse=None,
        uniform=False,
    ):
        if accept_sparse is None:
            accept_sparse = ["csc", "csr"]
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        self.uniform = uniform

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        views = self._validate_inputs(views)
        if self.uniform:
            weights = [np.ones((view.shape[1], self.latent_dims)) for view in views]
        else:
            weights = [
                self.random_state.normal(0, 1, size=(view.shape[1], self.latent_dims))
                for view in views
            ]
        self.weights = [weight / np.linalg.norm(weight, axis=0) for weight in weights]
        return self



import itertools
from typing import List, Union

import numpy as np
from scipy import linalg
from scipy.linalg import block_diag
from sklearn.utils.validation import check_random_state

from cca_zoo.utils import _process_parameter


class LinearSimulatedData:
    def __init__(
        self,
        view_features: List[int],
        latent_dims: int = 1,
        view_sparsity: List[Union[int, float]] = None,
        correlation: Union[List[float], float] = 0.99,
        structure: Union[str, List[str]] = None,
        sigma: Union[List[float], float] = None,
        decay: float = 0.5,
        positive=None,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        """

        Parameters
        ----------
        view_features : List[int]
            Number of features in each view
        latent_dims : int
            Number of latent dimensions
        view_sparsity : List[Union[int, float]]
            Sparsity of each view
        correlation : Union[List[float], float]
            Correlation between views
        structure : Union[str, List[str]]
            Structure of each view
        sigma : Union[List[float], float]
            Noise level of each view
        positive : None
            Unused
        random_state : Union[int, np.random.RandomState]
            Random state
        """
        self.view_features = view_features
        self.latent_dims = latent_dims
        self.correlation = correlation
        self.random_state = check_random_state(random_state)
        self.correlation = _process_parameter(
            "correlation", correlation, 0.99, latent_dims
        )
        # correlation must all be less than 1
        if np.any(np.abs(self.correlation) >= 1):
            raise ValueError("Magnitude of Correlation must be less than 1")
        self.structure = _process_parameter(
            "structure", structure, "identity", len(view_features)
        )
        self.view_sparsity = _process_parameter(
            "view_sparsity", view_sparsity, 1, len(view_features)
        )
        self.positive = _process_parameter(
            "positive", positive, False, len(view_features)
        )
        self.sigma = _process_parameter("sigma", sigma, 0.5, len(view_features))

        self.mean, covs, self.true_features = self._generate_covariance_matrices()
        joint_cov = self._generate_joint_covariance(covs)
        self.chol = np.linalg.cholesky(joint_cov)

    def _generate_covariance_matrix(self, view_p, view_structure, view_sigma):
        if view_structure == "identity":
            cov = np.eye(view_p)
        elif view_structure == "gaussian":
            cov = _generate_gaussian_cov(view_p, view_sigma)
        elif view_structure == "toeplitz":
            cov = _generate_toeplitz_cov(view_p, view_sigma)
        elif view_structure == "random":
            cov = _generate_random_cov(view_p, self.random_state)
        else:
            raise ValueError(f"Unknown structure {view_structure}")
        return cov

    def _generate_joint_covariance(self, covs):
        cov = block_diag(*covs)
        splits = np.concatenate(([0], np.cumsum(self.view_features)))
        for i, j in itertools.combinations(range(len(splits) - 1), 2):
            cross = np.zeros((self.view_features[i], self.view_features[j]))
            for _ in range(self.latent_dims):
                A = self.correlation[_] * np.outer(
                    self.true_features[i][:, _], self.true_features[j][:, _]
                )
                # Cross Bit
                cross += covs[i] @ A @ covs[j]
            cov[
                splits[i] : splits[i] + self.view_features[i],
                splits[j] : splits[j] + self.view_features[j],
            ] = cross
            cov[
                splits[j] : splits[j] + self.view_features[j],
                splits[i] : splits[i] + self.view_features[i],
            ] = cross.T
        return cov

    def _generate_covariance_matrices(self):
        mean = np.zeros(sum(self.view_features))
        covs = []
        true_features = []
        for view_p, sparsity, view_structure, view_positive, view_sigma in zip(
            self.view_features,
            self.view_sparsity,
            self.structure,
            self.positive,
            self.sigma,
        ):
            cov = self._generate_covariance_matrix(view_p, view_structure, view_sigma)
            weights = self.random_state.randn(view_p, self.latent_dims)
            if sparsity <= 1:
                sparsity = np.ceil(sparsity * view_p).astype("int")
            if sparsity < view_p:
                mask = np.stack(
                    (
                        np.concatenate(
                            ([0] * (view_p - sparsity), [1] * sparsity)
                        ).astype(bool),
                    )
                    * self.latent_dims,
                    axis=0,
                ).T
                mask = mask.flatten()
                self.random_state.shuffle(mask)
                mask = mask.reshape(weights.shape)
                weights = weights * mask
                if view_positive:
                    weights[weights < 0] = 0
            weights = _decorrelate_dims(weights, cov)
            weights /= np.sqrt(np.diag((weights.T @ cov @ weights)))
            true_features.append(weights)
            covs.append(cov)
        return mean, covs, true_features

    def sample(self, n: int):
        # check self.chol exists
        X = np.zeros((n, self.chol.shape[0]))
        for i in range(n):
            X[i, :] = self._chol_sample(self.mean, self.chol, self.random_state)
        views = np.split(X, np.cumsum(self.view_features)[:-1], axis=1)
        return views

    @staticmethod
    def _chol_sample(mean, chol, random_state):
        rng = check_random_state(random_state)
        return mean + chol @ rng.randn(mean.size)


def simple_simulated_data(
    n: int,
    view_features: List[int],
    view_sparsity: List[Union[int, float]] = None,
    eps: float = 0,
    transform=False,
    random_state=None,
):
    """
    Generate a simple simulated dataset with a single latent dimension

    Parameters
    ----------
    n : int
        Number of samples
    view_features :
        Number of features in each view
    view_sparsity : List[Union[int, float]]
        Sparsity of each view. If int, then number of non-zero features. If float, then proportion of non-zero features.
    eps : float
        Noise level
    transform : bool
        Whether to transform the data to be non-negative
    random_state : int, RandomState instance, default=None
        Controls the random seed in generating the data.

    Returns
    -------

    """
    random_state = check_random_state(random_state)
    z = random_state.randn(n)
    if transform:
        z = np.sin(z)
    views = []
    true_features = []
    view_sparsity = _process_parameter(
        "view_sparsity", view_sparsity, 0, len(view_features)
    )
    for p, sparsity in zip(view_features, view_sparsity):
        weights = random_state.normal(size=(p, 1))
        if sparsity <= 1:
            sparsity = np.ceil(sparsity * p).astype("int")
            weights[random_state.choice(np.arange(p), p - sparsity, replace=False)] = 0
        gaussian_x = random_state.normal(0, eps, size=(n, p)) * eps
        view = np.outer(z, weights)
        view += gaussian_x
        views.append(view / np.linalg.norm(view, axis=0))
        true_features.append(weights)
    return views, true_features


def _decorrelate_dims(up, cov):
    A = up.T @ cov @ up
    for k in range(1, A.shape[0]):
        up[:, k:] -= np.outer(up[:, k - 1], A[k - 1, k:] / A[k - 1, k - 1])
        A = up.T @ cov @ up
    return up


def _gaussian(x, mu, sig, dn):
    """
    Generate a gaussian covariance matrix

    :param x:
    :param mu:
    :param sig:
    :param dn:
    """
    return (
        np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))
        * dn
        / (np.sqrt(2 * np.pi) * sig)
    )


def _generate_gaussian_cov(p, sigma):
    x = np.linspace(-1, 1, p)
    x_tile = np.tile(x, (p, 1))
    mu_tile = np.transpose(x_tile)
    dn = 2 / (p - 1)
    cov = _gaussian(x_tile, mu_tile, sigma, dn)
    cov /= cov.max()
    return cov


def _generate_toeplitz_cov(p, sigma):
    c = np.arange(0, p)
    c = sigma**c
    cov = linalg.toeplitz(c, c)
    return cov


def _generate_random_cov(p, random_state):
    rng = check_random_state(random_state)
    cov_ = rng.randn(p, p)
    U, S, Vt = np.linalg.svd(cov_.T @ cov_)
    cov = U @ (1 + np.diag(rng.randn(p))) @ Vt
    return cov



import copy
import warnings
from abc import abstractmethod
from itertools import combinations
from typing import Union, Iterable

import numpy as np
from tqdm import tqdm

from cca_zoo.models._base import _BaseCCA
from cca_zoo.models._dummy import _DummyCCA
from cca_zoo.models._rcca import rCCA
from .._mcca import MCCA


class _BaseIterative(_BaseCCA):
    """
    A class used as the base for _iterative CCA methods i.e. those that optimize for each dimension one at a time with deflation.

    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        deflation="cca",
        max_iter: int = 100,
        initialization: Union[str, callable] = "random",
        tol: float = 1e-3,
        verbose=0,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=["csc", "csr"],
            random_state=random_state,
        )
        self.max_iter = max_iter
        self.initialization = initialization
        self.tol = tol
        self.deflation = deflation
        self.verbose = verbose

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        views = self._validate_inputs(views)
        self._check_params()
        self.weights = [
            np.zeros((view.shape[1], self.latent_dims), dtype=view.dtype)
            for view in views
        ]
        initializer = _default_initializer(
            views, self.initialization, self.random_state, self.latent_dims
        )
        initializer_scores = np.stack(initializer.fit_transform(views))
        residuals = copy.deepcopy(list(views))
        self.track = {"objective": {}}
        for k in (
            tqdm(range(self.latent_dims), desc="latent dimension")
            if self.verbose > 0
            else range(self.latent_dims)
        ):
            initial_weights = [w[:, k] for w in initializer.weights]
            weights, self.track["objective"][k] = self._fit(
                residuals, initializer_scores[:, :, k], initial_weights
            )
            for i, residual in enumerate(residuals):
                self.weights[i][:, k] = weights[i]
                residuals[i] = self._deflate(residuals[i], self.weights[i][:, k])
        return self

    def _fit(self, views, scores, weights):
        objective = []
        for t in (
            tqdm(range(self.max_iter), desc="inner loop iterations")
            if self.verbose > 1
            else range(self.max_iter)
        ):
            self._initialize(views)
            scores, weights = self._update(views, scores, weights)
            if np.isnan(scores).sum() > 0:
                warnings.warn(
                    f"Some scores are nan. Usually regularisation is too high."
                )
                break
            objective.append(self._objective(views, scores, weights))
            if t > 0 and self._early_stop(objective):
                break
        return weights, objective

    def _deflate(self, residual, weights):
        """
        Deflate view residual by CCA deflation (https://ars.els-cdn.com/content/image/1-s2.0-S0006322319319183-mmc1.pdf)

        :param residual: the current residual data matrix

        """
        score = residual @ weights
        if self.deflation == "cca":
            return residual - np.outer(score, score) @ residual / np.dot(score, score)
        elif self.deflation == "pls":
            return residual - np.outer(residual @ weights, weights)
        else:
            raise ValueError(f"deflation method {self.deflation} not implemented yet.")

    @abstractmethod
    def _update(self, views, scores, weights):
        return scores, weights

    def _objective(self, views, scores, weights) -> int:
        """
        Function used to calculate the objective function for the given. If we do not override then returns the covariance
         between projections

        :return:
        """
        # default objective is covariance
        obj = 0
        for score_i, score_j in combinations(scores, 2):
            obj += np.cov(score_i, score_j)[0, 1]
        return -obj.item()

    def _early_stop(self, objective) -> bool:
        # Some kind of early stopping
        if (
            np.abs(objective[-1] - objective[-2])
            / np.abs(objective[-1] + objective[-2])
            < self.tol
        ):
            return True
        else:
            return False

    def _initialize(self, views):
        pass


def _default_initializer(views, initialization, random_state, latent_dims):
    if initialization == "random":
        initializer = _DummyCCA(latent_dims, random_state=random_state, uniform=False)
    elif initialization == "uniform":
        initializer = _DummyCCA(latent_dims, random_state=random_state, uniform=True)
    elif initialization == "pls":
        initializer = rCCA(latent_dims, random_state=random_state, c=1)
    elif initialization == "cca":
        initializer = MCCA(latent_dims)
    else:
        raise ValueError(
            "Initialization {type} not supported. Pass a generator implementing this method"
        )
    return initializer


def _cosine_similarity(a, b):
    """
    Calculates the cosine similarity between vectors
    :param a: 1d numpy array
    :param b: 1d numpy array
    :return: cosine similarity
    """
    # https: // www.statology.org / cosine - similarity - python /
    return a.T @ b / (np.linalg.norm(a) * np.linalg.norm(b))



import numpy as np
from pyproximal.proximal.L1 import _softthreshold


def _bin_search(current, previous, current_val, previous_val, min_, max_):
    """
    Binary search helper function.

    Parameters
    ----------
    current : current parameter value
    previous :
    current_val :
    previous_val :
    min_ : minimum parameter value resulting in function value less than zero
    max_ : maximum parameter value resulting in function value greater than zero

    Returns
    -------
    new :
    current :
    min_ :
    max_ :

    """
    if previous_val is None:
        previous_val = current_val
    if current_val <= 0:
        if previous_val <= 0:
            new = (current + max_) / 2
        if previous_val > 0:
            new = (current + previous) / 2
        if current > min_:
            min_ = current
    if current_val > 0:
        if previous_val > 0:
            new = (current + min_) / 2
        if previous_val <= 0:
            new = (current + previous) / 2
        if current < max_:
            max_ = current
    return new, current, min_, max_


def _delta_search(w, c, init=0, tol=1e-9, max_iter=1000):
    """
    Searches for threshold delta such that the 1-norm of weights w is less than or equal to c
    Parameters
    ----------
    w : weights found by one power method iteration
    c : 1-norm threshold
    init : initial value for delta
    tol : tolerance for convergence

    Returns
    -------
    updated weights

    """
    # First normalise the weights unit length
    w = w / np.linalg.norm(w, 2)
    converged = False
    min_ = 0
    max_ = 10
    current = init
    previous = current
    previous_val = 0
    i = 0
    while not converged:
        i += 1
        coef = _softthreshold(w, current)
        if np.linalg.norm(coef) > 0:
            coef /= np.linalg.norm(coef)
        current_val = c - np.linalg.norm(coef, 1)
        current, previous, min_, max_ = _bin_search(
            current, previous, current_val, previous_val, min_, max_
        )
        if (np.abs(current_val) < tol) or i == max_iter:
            converged = True
        previous_val = current_val
    return coef


def support_threshold(data, support, **kwargs):
    idx = np.argpartition(data.ravel(), data.shape[0] - support)
    data[idx[:-support]] = 0
    return data



from . import architectures
from . import callbacks
from . import objectives
from ._discriminative import (
    DCCA,
    DCCA_NOI,
    BarlowTwins,
    DCCA_SDL,
    DTCCA,
    DCCA_EigenGame,
)
from ._generative import DVCCA, SplitAE, DCCAE

__all__ = [
    "DCCA",
    "DCCAE",
    "DCCA_NOI",
    "DCCA_SDL",
    "DVCCA",
    "BarlowTwins",
    "DTCCA",
    "DCCA_EigenGame",
    "SplitAE",
    "architectures",
    "objectives",
    "callbacks",
]

classes = [
    "DCCA",
    "DCCAE",
    "DCCA_NOI",
    "DCCA_SDL",
    "DVCCA",
    "BarlowTwins",
    "DTCCA",
    "SplitAE",
    "DCCA_EigenGame",
]



import torch

from ._dcca import DCCA


class DCCA_EigenGame(DCCA):
    """

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(self, latent_dims: int, encoders=None, r: float = 0, **kwargs):
        super().__init__(latent_dims=latent_dims, encoders=encoders, **kwargs)
        self.r = r

    def forward(self, views, **kwargs):
        z = []
        for i, encoder in enumerate(self.encoders):
            z.append(encoder(views[i]))
        return z

    def loss(self, views, **kwargs):
        z = self(views)
        A, B = self.get_AB(z)
        rewards = 2 * torch.trace(A)
        penalties = torch.trace(A @ B)
        return {
            "objective": -rewards.sum() + penalties,
            "rewards": rewards.sum(),
            "penalties": penalties,
        }

    def get_AB(self, z):
        # sum the pairwise covariances between each z and all other zs
        A = torch.zeros(self.latent_dims, self.latent_dims, device=z[0].device)
        B = torch.zeros(self.latent_dims, self.latent_dims, device=z[0].device)
        for i, zi in enumerate(z):
            for j, zj in enumerate(z):
                if i == j:
                    B += torch.cov(zi.T)
                A += torch.cov(torch.hstack((zi, zj)).T)[
                    self.latent_dims :, : self.latent_dims
                ]
        return A / len(z), B / len(z)



"""
More than 2 views
===========================

This will compare MCCA, GCCA, TCCA for linear models with more than 2 views
"""
import numpy as np

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.models import MCCA, GCCA, TCCA, KCCA, KGCCA, KTCCA, SCCA_PMD

"""
Data
-----
"""
# %%
np.random.seed(42)
n = 30
p = 3
q = 3
r = 3
latent_dims = 1
cv = 3

(X, Y, Z) = LinearSimulatedData(
    view_features=[p, q, r], latent_dims=latent_dims, correlation=[0.9]
).sample(n)

# %%
# Eigendecomposition-Based Methods
# ---------------------------------

# %%
# Linear
# ^^^^^^^^

# %%
mcca = MCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))

# %%
gcca = GCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))

# %%
# Kernel
# ^^^^^^^^

# %%
kcca = KCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))

# %%
kgcca = KGCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))

# %%
# Iterative Methods
# ^^^^^^^^^^^^^^^^^^

# Most of the _iterative methods can also use multiple views e.g.

pmd = (
    SCCA_PMD(latent_dims=latent_dims, tau=0.1, tol=1e-5).fit((X, Y, X)).score((X, Y, Z))
)

# %%
# Higher Order Correlations
# -------------------------

# %%
# Tensor CCA finds higher order correlations so scores are not comparable (but TCCA is equivalent for 2 views)
tcca = TCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))

# %%
ktcca = KTCCA(latent_dims=latent_dims).fit((X, Y, X)).score((X, Y, Z))



from typing import Iterable, Union

import numpy as np
from scipy.linalg import block_diag
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.utils.validation import check_is_fitted

from cca_zoo.models._rcca import rCCA
from cca_zoo.utils.check_values import _process_parameter, _check_views


class MCCA(rCCA):
    r"""
    A class used to fit MCCA model. For more than 2 views, MCCA optimizes the sum of pairwise correlations.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} w_i^TX_i^TX_jw_j  \}\\

        \text{subject to:}

        (1-c_i)w_i^TX_i^TX_iw_i+c_iw_i^Tw_i=1

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state, by default None
    c : Union[Iterable[float], float], optional
        Regularisation parameter, by default None
    eps : float, optional
        Small value to add to the diagonal of the regularisation matrix, by default 1e-9


    References
    ----------
    Kettenring, Jon R. "Canonical analysis of several sets of variables." Biometrika 58.3 (1971): 433-451.

    Examples
    --------
    >>> from cca_zoo.models import MCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = MCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    array([0.97200847])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        eps=1e-9,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=["csc", "csr"],
            random_state=random_state,
        )
        self.c = c
        self.eps = eps

    def _weights(self, eigvals, eigvecs, views):
        self.weights = [
            eigvecs[split : self.splits[i + 1]]
            for i, split in enumerate(self.splits[:-1])
        ]

    def _setup_evp(self, views: Iterable[np.ndarray], **kwargs):
        all_views = np.hstack(views)
        C = np.cov(all_views, rowvar=False)
        # Can regularise by adding to diagonal
        D = block_diag(
            *[
                (1 - self.c[i]) * np.cov(view, rowvar=False)
                + self.c[i] * np.eye(view.shape[1])
                for i, view in enumerate(views)
            ]
        )
        C -= block_diag(*[np.cov(view, rowvar=False) for view in views])
        D_smallest_eig = min(0, np.linalg.eigvalsh(D).min()) - self.eps
        D = D - D_smallest_eig * np.eye(D.shape[0])
        self.splits = np.cumsum([0] + [view.shape[1] for view in views])
        return C / len(views), D / len(views)

    def _get_weights(self, eigvals, eigvecs, views):
        self.weights = [
            eigvecs[split : self.splits[i + 1]]
            for i, split in enumerate(self.splits[:-1])
        ]

    def _more_tags(self):
        return {"multiview": True}


class KCCA(MCCA):
    r"""
    A class used to fit KCCA model.

    .. math::

        \alpha_{opt}=\underset{\alpha}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \alpha_i^TK_i^TK_j\alpha_j  \}\\

        \text{subject to:}

        c_i\alpha_i^TK_i\alpha_i + (1-c_i)\alpha_i^TK_i^TK_i\alpha_i=1

    Examples
    --------
    >>> from cca_zoo.models import KCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> X3 = rng.random((10,5))
    >>> model = KCCA()
    >>> model.fit((X1,X2,X3)).score((X1,X2,X3))
    array([0.96893666])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
        eps=1e-3,
        kernel: Iterable[Union[float, callable]] = None,
        gamma: Iterable[float] = None,
        degree: Iterable[float] = None,
        coef0: Iterable[float] = None,
        kernel_params: Iterable[dict] = None,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
        )
        self.kernel_params = kernel_params
        self.gamma = gamma
        self.coef0 = coef0
        self.kernel = kernel
        self.degree = degree
        self.c = c
        self.eps = eps

    def _check_params(self):
        self.kernel = _process_parameter("kernel", self.kernel, "linear", self.n_views)
        self.gamma = _process_parameter("gamma", self.gamma, None, self.n_views)
        self.coef0 = _process_parameter("coef0", self.coef0, 1, self.n_views)
        self.degree = _process_parameter("degree", self.degree, 1, self.n_views)
        self.c = _process_parameter("c", self.c, 0, self.n_views)

    def _get_kernel(self, view, X, Y=None):
        if callable(self.kernel[view]):
            params = self.kernel_params[view] or {}
        else:
            params = {
                "gamma": self.gamma[view],
                "degree": self.degree[view],
                "coef0": self.coef0[view],
            }
        return pairwise_kernels(
            X, Y, metric=self.kernel[view], filter_params=True, **params
        )

    def _setup_evp(self, views: Iterable[np.ndarray], **kwargs):
        self.train_views = views
        kernels = [self._get_kernel(i, view) for i, view in enumerate(self.train_views)]
        C = np.cov(np.hstack(kernels), rowvar=False)
        # Can regularise by adding to diagonal
        D = block_diag(
            *[
                (1 - self.c[i]) * np.cov(kernel, rowvar=False) + self.c[i] * kernel
                for i, kernel in enumerate(kernels)
            ]
        )
        C -= block_diag(*[np.cov(kernel, rowvar=False) for kernel in kernels]) - D
        D_smallest_eig = min(0, np.linalg.eigvalsh(D).min()) - self.eps
        D = D - D_smallest_eig * np.eye(D.shape[0])
        self.splits = np.cumsum([0] + [kernel.shape[1] for kernel in kernels])
        return C / len(views), D / len(views)

    def transform(self, views: np.ndarray, **kwargs):
        check_is_fitted(self, attributes=["weights"])
        _check_views(views)
        views = [self.scalers[i].transform(view) for i, view in enumerate(views)]
        Ktest = [
            self._get_kernel(i, self.train_views[i], Y=view)
            for i, view in enumerate(views)
        ]
        transformed_views = [
            kernel.T @ self.weights[i] for i, kernel in enumerate(Ktest)
        ]
        return transformed_views



import warnings
from typing import Iterable

import numpy as np

from cca_zoo.models._mcca import MCCA
from ..utils import _process_parameter


class GRCCA(MCCA):
    """
    Grouped Regularized Canonical Correlation Analysis

    Parameters
    ----------
    latent_dims : int, default=1
        Number of latent dimensions to use
    scale : bool, default=True
        Whether to scale the data to unit variance
    centre : bool, default=True
        Whether to centre the data
    copy_data : bool, default=True
        Whether to copy the data
    random_state : int, default=None
        Random state for initialisation
    eps : float, default=1e-3
        Tolerance for convergence
    c : float, default=0
        Regularization parameter for the group means
    mu : float, default=0
        Regularization parameter for the group sizes


    References
    ----------
    Tuzhilina, Elena, Leonardo Tozzi, and Trevor Hastie. "Canonical correlation analysis in high dimensions with structured regularization." Statistical Modelling (2021): 1471082X211041033.
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        eps=1e-3,
        c: float = 0,
        mu: float = 0,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
            eps=eps,
            c=c,
        )
        self.mu = mu

    def _check_params(self):
        self.mu = _process_parameter("mu", self.mu, 0, self.n_views)
        self.c = _process_parameter("c", self.c, 0, self.n_views)

    def fit(self, views: Iterable[np.ndarray], y=None, feature_groups=None, **kwargs):
        """
        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        feature_groups : list/tuple of integer numpy arrays or array likes with dimensions (,view shape)
        kwargs: any additional keyword arguments required by the given model

        """
        if feature_groups is None:
            warnings.warn(f"No feature groups provided, using all features")
            feature_groups = [np.ones(view.shape[1], dtype=int) for view in views]
        for feature_group in feature_groups:
            assert np.issubdtype(
                feature_group.dtype, np.integer
            ), "feature groups must be integers"
        views = self._validate_inputs(views)
        self._check_params()
        views, idxs = self._preprocess(views, feature_groups)
        C, D = self._setup_evp(views, **kwargs)
        eigvals, eigvecs = self._solve_evp(C, D)
        self._weights(eigvals, eigvecs, views)
        self._transform_weights(views, feature_groups)
        return self

    def _preprocess(self, views, feature_groups):
        views, idxs = list(
            zip(
                *[
                    self._process_view(view, group, mu, c)
                    for view, group, mu, c in zip(
                        views, feature_groups, self.mu, self.c
                    )
                ]
            )
        )
        return views, idxs

    @staticmethod
    def _process_view(view, group, mu, c):
        if c > 0:
            ids, unique_inverse, unique_counts, group_means = _group_mean(view, group)
            if mu == 0:
                mu = 1
                idx = view.shape[1] - 1
            else:
                idx = view.shape[1] + group_means.shape[1] - 1
            view_1 = (view - group_means[:, unique_inverse]) / c
            view_2 = group_means / np.sqrt(mu / unique_counts)
            return np.hstack((view_1, view_2)), idx
        else:
            return view, view.shape[1] - 1

    def _transform_weights(self, views, groups):
        for i, (view, group) in enumerate(zip(views, groups)):
            if self.c[i] > 0:
                weights_1 = self.weights[i][: len(group)]
                weights_2 = self.weights[i][len(group) :]
                ids, unique_inverse, unique_counts, group_means = _group_mean(
                    weights_1.T, group
                )
                weights_1 = (weights_1 - group_means[:, unique_inverse].T) / self.c[i]
                if self.mu[i] == 0:
                    mu = 1
                else:
                    mu = self.mu[i]
                weights_2 = weights_2 / np.sqrt(
                    mu * np.expand_dims(unique_counts, axis=1)
                )
                self.weights[i] = weights_1 + weights_2[group]

    def _more_tags(self):
        return {"multiview": True}


def _group_mean(view, group):
    ids, unique_inverse, unique_counts = np.unique(
        group, return_inverse=True, return_counts=True
    )
    group_means = np.array([view[:, group == id].mean(axis=1) for id in ids]).T
    return ids, unique_inverse, unique_counts, group_means



__all__ = [
    "data",
    "model_selection",
    "models",
    "plotting",
]

# if can import deepmodels add to all
try:
    import cca_zoo.deepmodels

    __all__.append("deepmodels")
except ModuleNotFoundError:
    pass
try:
    import cca_zoo.probabilisticmodels

    __all__.append("probabilisticmodels")
except ModuleNotFoundError:
    pass



from typing import Union, Iterable

import numpy as np

from cca_zoo.models._iterative._base import _BaseIterative
from cca_zoo.models._search import _softthreshold
from cca_zoo.utils import _process_parameter, _check_converged_weights


class SCCA_Parkhomenko(_BaseIterative):
    r"""
    Fits a sparse CCA (penalized CCA) model for 2 or more views.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \} + c_i\|w_i\|\\

        \text{subject to:}

        w_i^Tw_i=1

    References
    ----------
    Parkhomenko, Elena, David Tritchler, and Joseph Beyene. "Sparse canonical correlation analysis with application to genomic data integration." Statistical applications in genetics and molecular biology 8.1 (2009).

    Examples
    --------
    >>> from cca_zoo.models import SCCA_Parkhomenko
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_Parkhomenko(tau=[0.001,0.001],random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.81803527])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        deflation="cca",
        tau: Union[Iterable[float], float] = None,
        max_iter: int = 100,
        initialization: Union[str, callable] = "pls",
        tol: float = 1e-3,
        verbose=0,
    ):
        self.tau = tau
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            verbose=verbose,
        )

    def _check_params(self):
        self.tau = _process_parameter("tau", self.tau, 0.0001, self.n_views)
        if any(tau <= 0 for tau in self.tau):
            raise (
                "All regularisation parameters should be above 0. " f"tau=[{self.tau}]"
            )

    def _update(self, views, scores, weights):
        for view_index, view in enumerate(views):
            # mask off the current view and sum the rest
            targets = np.ma.array(scores, mask=False)
            targets.mask[view_index] = True
            w = views[view_index].T @ targets.sum(axis=0).filled()
            w /= np.linalg.norm(w)
            _check_converged_weights(w, view_index)
            w = _softthreshold(w, self.tau[view_index] / 2)
            weights[view_index] = w / np.linalg.norm(w)
            _check_converged_weights(w, view_index)
            scores[view_index] = views[view_index] @ weights[view_index]
        return scores, weights

    def _more_tags(self):
        return {"multiview": True}



import numpy as np
from torch.utils.data import Dataset, DataLoader


class NumpyDataset(Dataset):
    """
    Class that turns numpy arrays into a torch dataset
    """

    def __init__(self, views, labels=None):
        """

        :param views: list/tuple of numpy arrays or array likes with the same number of rows (samples)
        """
        self.labels = labels
        self.views = [view.astype(np.float32) for view in views]

    def __len__(self):
        return len(self.views[0])

    def __getitem__(self, index):
        views = [view[index] for view in self.views]
        if self.labels is not None:
            label = self.labels[index]
            return {"views": views, "label": label}
        else:
            return {"views": views}


def check_dataset(dataset):
    """
    Checks that a custom dataset returns a dictionary with a "views" key containing a list of tensors

    Parameters
    ----------
    dataset: torch.utils.data.Dataset

    Returns
    -------

    """
    dataloader = DataLoader(
        dataset,
    )
    for batch in dataloader:
        if "views" not in batch:
            raise ValueError(
                "The dataset must return a dictionary with a 'views' key containing a list of tensors"
            )
        else:
            break


def get_dataloaders(
    dataset,
    val_dataset=None,
    batch_size=None,
    val_batch_size=None,
    drop_last=True,
    val_drop_last=False,
    shuffle_train=False,
    pin_memory=True,
    num_workers=0,
    persistent_workers=True,
):
    """
    A utility function to allow users to quickly get hold of the dataloaders required by pytorch lightning

    :param dataset: A CCA dataset used for training
    :param val_dataset: An optional CCA dataset used for validation
    :param batch_size: batch size of train loader
    :param val_batch_size: batch size of val loader
    :param num_workers: number of workers used
    :param pin_memory: pin memory used by pytorch - True tends to speed up training
    :param shuffle_train: whether to shuffle training data
    :param val_drop_last: whether to drop the last incomplete batch from the validation data
    :param drop_last: whether to drop the last incomplete batch from the train data
    :param persistent_workers: whether to keep workers alive after dataloader is destroyed

    """
    if num_workers == 0:
        persistent_workers = False
    if batch_size is None:
        batch_size = len(dataset)
    train_dataloader = DataLoader(
        dataset,
        batch_size=batch_size,
        drop_last=drop_last,
        num_workers=num_workers,
        pin_memory=pin_memory,
        shuffle=shuffle_train,
        persistent_workers=persistent_workers,
    )
    if val_dataset:
        if val_batch_size is None:
            val_batch_size = len(val_dataset)
        val_dataloader = DataLoader(
            val_dataset,
            batch_size=val_batch_size,
            drop_last=val_drop_last,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )
        return train_dataloader, val_dataloader
    return train_dataloader



import warnings
from typing import Union, Iterable

import numpy as np

from cca_zoo.models._search import _delta_search
from cca_zoo.utils import _process_parameter, _check_converged_weights
from ._pls_als import PLS_ALS


class SCCA_PMD(PLS_ALS):
    r"""
    Fits a Sparse CCA (Penalized Matrix Decomposition) model for 2 or more views.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{ w_1^TX_1^TX_2w_2  \}\\

        \text{subject to:}

        w_i^Tw_i=1

        \|w_i\|<=c_i

    Parameters
    ----------
    latent_dims : int, default=1
        Number of latent dimensions to use in the model.
    scale : bool, default=True
        Whether to scale the data to unit variance.
    centre : bool, default=True
        Whether to centre the data to have zero mean.
    copy_data : bool, default=True
        Whether to copy the data or overwrite it.
    random_state : int, default=None
        Random seed for initialisation.
    deflation : str, default="cca"
        Deflation method to use. Options are "cca" and "pmd".
    tau : float or list of floats, default=None
        Regularisation parameter. If a single float is given, the same value is used for all views.
        If a list of floats is given, the values are used for each view respectively.
        If None, the value is set to 1.
    max_iter : int, default=100
        Maximum number of iterations to run.
    initialization : str or callable, default="pls"
        Method to use for initialisation. Options are "pls" and "random".
    tol : float, default=1e-9
        Tolerance for convergence.
    positive : bool or list of bools, default=False
        Whether to constrain the weights to be positive.
    verbose : int, default=0
        Verbosity level. 0 is silent, 1 prints progress.


    References
    ----------
    Witten, Daniela M., Robert Tibshirani, and Trevor Hastie. "A penalized matrix decomposition, with applications to sparse principal components and canonical correlation analysis." Biostatistics 10.3 (2009): 515-534.

    Examples
    --------
    >>> from cca_zoo.models import SCCA_PMD
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_PMD(tau=[1,1],random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.81796873])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        deflation="cca",
        tau: Union[Iterable[float], float] = None,
        max_iter: int = 100,
        initialization: Union[str, callable] = "pls",
        tol: float = 1e-3,
        positive: Union[Iterable[bool], bool] = None,
        verbose=0,
    ):
        self.tau = tau
        self.positive = positive
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            verbose=verbose,
        )

    def _check_params(self):
        if self.tau is None:
            warnings.warn(
                "tau parameter not set. Setting to tau=1 i.e. maximum regularisation of l1 norm"
            )
        self.tau = _process_parameter("tau", self.tau, 1, self.n_views)
        if any(tau < 0 or tau > 1 for tau in self.tau):
            raise ValueError(
                "All regularisation parameters should be between 0 and 1 "
                f"1. tau=[{self.tau}]"
            )
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views
        )

    def _initialize(self, views):
        shape_sqrts = [np.sqrt(view.shape[1]) for view in views]
        self.t = [max(1, x * y) for x, y in zip(self.tau, shape_sqrts)]

    def _update(self, views, scores, weights):
        # Update each view using loop update function
        for view_index, view in enumerate(views):
            targets = np.ma.array(scores, mask=False)
            targets.mask[view_index] = True
            weights[view_index] = (
                views[view_index].T
                @ targets.sum(axis=0).filled()
                / (views[0].shape[0] - 1)
            )
            weights[view_index] = _delta_search(
                weights[view_index],
                self.t[view_index],
                tol=self.tol,
            )
            _check_converged_weights(weights[view_index], view_index)
            scores[view_index] = views[view_index] @ weights[view_index]
        return scores, weights



"""
Sparse CCA Methods
===========================

This script shows how regularised methods can be used to extract sparse solutions to the CCA problem
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.model_selection import GridSearchCV
from cca_zoo.models import (
    SCCA_PMD,
    SCCA_IPLS,
    ElasticCCA,
    CCA,
    PLS,
    SCCA_Span,
    AltMaxVar,
)


def plot_true_weights_coloured(ax, weights, true_weights, title=""):
    ind = np.arange(len(true_weights))
    mask = np.squeeze(true_weights == 0)
    ax.scatter(ind[~mask], weights[~mask], c="b")
    ax.scatter(ind[mask], weights[mask], c="r")
    ax.set_title(title)


def plot_model_weights(wx, wy, tx, ty, title=""):
    fig, axs = plt.subplots(2, 2, sharex=True, sharey=True)
    plot_true_weights_coloured(axs[0, 0], tx, tx, title="true x weights")
    plot_true_weights_coloured(axs[0, 1], ty, ty, title="true y weights")
    plot_true_weights_coloured(axs[1, 0], wx, tx, title="model x weights")
    plot_true_weights_coloured(axs[1, 1], wy, ty, title="model y weights")
    plt.tight_layout()
    fig.suptitle(title)
    plt.show()


# %%
# Data
# -----
np.random.seed(42)
n = 200
p = 100
q = 100
view_1_sparsity = 0.1
view_2_sparsity = 0.1
latent_dims = 1

data = LinearSimulatedData(
    view_features=[p, q],
    latent_dims=latent_dims,
    view_sparsity=[view_1_sparsity, view_2_sparsity],
    correlation=[0.9],
)
(X, Y) = data.sample(n)

tx = data.true_features[0] / np.sqrt(n)
ty = data.true_features[1] / np.sqrt(n)

# %% CCA
cca = CCA().fit([X, Y])
plot_model_weights(cca.weights[0], cca.weights[1], tx, ty, title="CCA")

# %% PLS
pls = PLS().fit([X, Y])
plot_model_weights(pls.weights[0], pls.weights[1], tx, ty, title="PLS")

# %% PMD
tau1 = [0.1, 0.3, 0.7, 0.9]
tau2 = [0.1, 0.3, 0.7, 0.9]
param_grid = {"tau": [tau1, tau2]}
pmd = GridSearchCV(SCCA_PMD(), param_grid=param_grid, cv=3, verbose=True).fit([X, Y])

# %%
plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(pmd.best_estimator_.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")
# %%
plot_model_weights(
    pmd.best_estimator_.weights[0], pmd.best_estimator_.weights[1], tx, ty, title="PMD"
)

# %%
pd.DataFrame(pmd.cv_results_)

# %% IPLS
scca = SCCA_IPLS(tau=[1e-2, 1e-2]).fit([X, Y])
plot_model_weights(scca.weights[0], scca.weights[1], tx, ty, title="SCCA_IPLS")

plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(scca.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")

scca_pos = SCCA_IPLS(tau=[1e-2, 1e-2], positive=[True, True]).fit([X, Y])
plot_model_weights(
    scca_pos.weights[0], scca_pos.weights[1], tx, ty, title="SCCA_IPLS (Positive)"
)

plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(scca_pos.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")

elasticcca = ElasticCCA(alpha=[1e-2, 1e-2], l1_ratio=[0.5, 0.5]).fit([X, Y])
plot_model_weights(
    elasticcca.weights[0], elasticcca.weights[1], tx, ty, title="ELastic CCA"
)

plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(elasticcca.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")

altmaxvar = AltMaxVar(tau=[1e-5, 1e-5]).fit([X, Y])
plot_model_weights(
    altmaxvar.weights[0], altmaxvar.weights[1], tx, ty, title="AltMaxVar"
)

plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(altmaxvar.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")

spancca = SCCA_Span(tau=[10, 10], max_iter=2000, rank=20).fit([X, Y])
plot_model_weights(spancca.weights[0], spancca.weights[1], tx, ty, title="Span CCA")

plt.figure()
plt.title("Objective Convergence")
plt.plot(np.array(spancca.track["objective"][0]))
plt.ylabel("Objective")
plt.xlabel("#iterations")



import warnings
from typing import Iterable

import numpy as np
from scipy.linalg import block_diag

from cca_zoo.models._mcca import MCCA


class PRCCA(MCCA):
    """
    Partially Regularized Canonical Correlation Analysis


    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state for reproducibility, by default None
    eps : float, optional
        Tolerance for convergence, by default 1e-3
    c : Union[Iterable[float], float], optional
        Regularisation parameter, by default None

    References
    ----------
    Tuzhilina, Elena, Leonardo Tozzi, and Trevor Hastie. "Canonical correlation analysis in high dimensions with structured regularization." Statistical Modelling (2021): 1471082X211041033.
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        eps=1e-3,
        c=0,
    ):
        """
        Parameters
        ----------
        c : Union[Iterable[float], float], optional
            Regularisation parameter, by default None
        eps : float, optional
            Tolerance for convergence, by default 1e-3
        """
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
            eps=eps,
            c=c,
        )

    def fit(self, views: Iterable[np.ndarray], y=None, idxs=None, **kwargs):
        """
        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        idxs : list/tuple of integers indicating which features from each view are the partially regularised features
        kwargs: any additional keyword arguments required by the given model

        """
        if idxs is None:
            warnings.warn(f"No idxs provided, using all features")
            idxs = [np.arange(views[0].shape[1], dtype=int)] * len(views)
        for idx in idxs:
            assert np.issubdtype(
                idx.dtype, np.integer
            ), "feature groups must be integers"
        views = self._validate_inputs(views)
        self._check_params()
        views = self._pretransform(views, idxs)
        self.weights = (
            MCCA(latent_dims=self.latent_dims, c=self.c, scale=False, centre=False)
            .fit(views, idxs=idxs)
            .weights
        )
        self._transform_weights(views, idxs=idxs)
        return self

    def _pretransform(self, views, idxs):
        X_1 = [view[:, idx] for view, idx in zip(views, idxs)]
        self.p = [X_i.shape[1] for X_i in X_1]
        X_2 = [np.delete(view, idx, axis=1) for view, idx in zip(views, idxs)]
        self.B = [np.linalg.pinv(X_2) @ X_1 for X_1, X_2 in zip(X_1, X_2)]
        X_1 = [X_1 - X_2 @ B for X_1, X_2, B in zip(X_1, X_2, self.B)]
        views = [np.hstack((X_1, X_2)) for X_1, X_2 in zip(X_1, X_2)]
        return views

    def _setup_evp(self, views: Iterable[np.ndarray], idxs=None, **kwargs):
        all_views = np.concatenate(views, axis=1)
        C = all_views.T @ all_views / self.n
        penalties = [np.zeros((view.shape[1])) for view in views]
        for i, idx in enumerate(idxs):
            penalties[i][idx] = self.c[i]
        D = block_diag(
            *[
                (1 - self.c[i]) * (m.T @ m) / self.n + np.diag(penalties[i])
                for i, m in enumerate(views)
            ]
        )
        C -= block_diag(*[view.T @ view / self.n for view in views])
        D_smallest_eig = min(0, np.linalg.eigvalsh(D).min()) - self.eps
        D = D - D_smallest_eig * np.eye(D.shape[0])
        self.splits = np.cumsum([0] + [view.shape[1] for view in views])
        return C, D

    def _transform_weights(self, views, idxs=None):
        for i, idx in enumerate(idxs):
            alpha_1 = self.weights[i][idx]
            alpha_2 = np.delete(self.weights[i], idx, axis=0)
            alpha_2 -= self.B[i] @ alpha_1
            mask = np.ones(self.weights[i].shape[0], dtype=bool)
            mask[idx] = False
            self.weights[i][mask] = alpha_2

    def _more_tags(self):
        return {"multiview": True}



from ._dcca import DCCA
from ._dcca_barlow_twins import BarlowTwins
from ._dcca_eigengame import DCCA_EigenGame
from ._dcca_gha import DCCA_GHA
from ._dcca_noi import DCCA_NOI
from ._dcca_sdl import DCCA_SDL
from ._dtcca import DTCCA



import torch

from cca_zoo.deepmodels import objectives
from cca_zoo.deepmodels._base import _BaseDeep
from cca_zoo.deepmodels.callbacks import CorrelationCallback
from cca_zoo.models import MCCA
from cca_zoo.models._base import _BaseCCA


class DCCA(_BaseDeep, _BaseCCA):
    """
    A class used to fit a DCCA model.

    References
    ----------
    Andrew, Galen, et al. "Deep canonical correlation analysis." International conference on machine learning. PMLR, 2013.

    """

    def __init__(
        self,
        latent_dims: int,
        objective=objectives.MCCA,
        encoders=None,
        r: float = 0,
        eps: float = 1e-5,
        **kwargs,
    ):
        super().__init__(latent_dims=latent_dims, **kwargs)
        self.encoders = torch.nn.ModuleList(encoders)
        self.objective = objective(latent_dims, r=r, eps=eps)

    def forward(self, views, **kwargs):
        z = []
        for i, encoder in enumerate(self.encoders):
            z.append(encoder(views[i]))
        return z

    def loss(self, views, **kwargs):
        z = self(views)
        return {"objective": self.objective.loss(z)}

    def pairwise_correlations(
        self,
        loader: torch.utils.data.DataLoader,
        train=False,
    ):
        """
        Calculates correlation for entire batch from dataloader

        :param loader: a dataloader that matches the structure of that used for training
        :param train: whether to fit final linear transformation
        :return: by default returns the average pairwise correlation in each dimension (for 2 views just the correlation)
        """
        return _BaseCCA.pairwise_correlations(self, loader, train=train)

    def score(self, loader: torch.utils.data.DataLoader, **kwargs):
        """
        Returns average correlation in each dimension (averages over all pairs for multiview)

        :param **kwargs:
        :type **kwargs:
        :param loader: a dataloader that matches the structure of that used for training
        :param train: whether to fit final linear transformation
        """
        z = self.transform(loader)
        return MCCA(self.latent_dims).fit(z).score(z)

    def configure_callbacks(self):
        return [CorrelationCallback()]



from typing import Union

import numpy as np

from cca_zoo.models._iterative._base import _BaseIterative


class PLS_ALS(_BaseIterative):
    r"""
    A class used to fit a PLS model to two or more views of data.

    Fits a partial least squares model with CCA deflation by NIPALS algorithm

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2\}\\

        \text{subject to:}

        w_i^Tw_i=1

    Can also be used with more than two views

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state for reproducibility, by default None
    max_iter : int, optional
        Maximum number of iterations, by default 100
    initialization : Union[str, callable], optional
        Initialization method, by default "random"
    tol : float, optional
        Tolerance for convergence, by default 1e-9
    verbose : int, optional
        Verbosity level, by default 0

    Examples
    --------

    >>> from cca_zoo.models import PLS
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = PLS_ALS(random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.81796854])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        max_iter: int = 100,
        initialization: Union[str, callable] = "random",
        tol: float = 1e-3,
        deflation="pls",
        verbose=0,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            deflation=deflation,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            verbose=verbose,
        )

    def _update(self, views, scores, weights):
        # Update each view using loop update function
        for view_index, view in enumerate(views):
            # mask off the current view and sum the rest
            targets = np.ma.array(scores, mask=False)
            targets.mask[view_index] = True
            weights[view_index] = (
                views[view_index].T
                @ targets.sum(axis=0).filled()
                / (views[0].shape[0] - 1)
            )
            weights[view_index] /= np.linalg.norm(weights[view_index])
            scores[view_index] = views[view_index] @ np.squeeze(
                np.array(weights[view_index])
            )
        return scores, weights

    def _more_tags(self):
        return {"multiview": True}



from itertools import combinations
from typing import Union, Iterable

import numpy as np

from cca_zoo.models._iterative._base import _BaseIterative
from cca_zoo.models._search import _delta_search, support_threshold
from cca_zoo.utils import _process_parameter


class SWCCA(_BaseIterative):
    r"""
    A class used to fit SWCCA model

    References
    ----------
    .. Wenwen, M. I. N., L. I. U. Juan, and Shihua Zhang. "Sparse Weighted Canonical Correlation Analysis." Chinese Journal of Electronics 27.3 (2018): 459-466.

    Examples
    --------
    >>> from cca_zoo.models import SWCCA
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SWCCA(regularisation='l0',tau=[2, 2], sample_support=5, random_state=0)
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.61620969])
    """

    def __init__(
        self,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        max_iter: int = 500,
        initialization: str = "random",
        tol: float = 1e-3,
        regularisation="l0",
        tau: Union[Iterable[Union[float, int]], Union[float, int]] = None,
        sample_support=None,
        positive=False,
        verbose=0,
    ):
        self.tau = tau
        self.sample_support = sample_support
        if regularisation == "l0":
            self.update = support_threshold
        elif regularisation == "l1":
            self.update = _delta_search
        self.regularisation = regularisation
        self.positive = positive
        super().__init__(
            latent_dims=1,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            verbose=verbose,
        )

    def _check_params(self):
        if self.sample_support is None:
            self.sample_support = self.n
        self.tau = _process_parameter("tau", self.tau, 2, self.n_views)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views
        )

    def _initialize(self, views):
        self.sample_weights = np.ones(self.n)
        self.sample_weights /= np.linalg.norm(self.sample_weights)

    def _update(self, views, scores, weights):
        # Update each view using loop update function
        for view_index, view in enumerate(views):
            targets = np.ma.array(scores, mask=False)
            targets.mask[view_index] = True
            weights[view_index] = (
                views[view_index] * self.sample_weights[:, np.newaxis]
            ).T @ targets.sum(axis=0).filled()
            weights[view_index] = self.update(
                weights[view_index],
                self.tau[view_index],
                positive=self.positive[view_index],
            )
            weights[view_index] /= np.linalg.norm(weights[view_index])
            if view_index == self.n_views - 1:
                self.sample_weights = self._update_sample_weights(scores)
            scores[view_index] = views[view_index] @ weights[view_index]
        return scores, weights

    def _update_sample_weights(self, scores):
        w = scores.prod(axis=0)
        sample_weights = support_threshold(w, self.sample_support)
        sample_weights /= np.linalg.norm(sample_weights)
        return sample_weights

    def _objective(self, views, scores, weights) -> int:
        # default objective is correlation
        obj = 0
        for score_i, score_j in combinations(scores, 2):
            obj += (score_i * self.sample_weights).T @ score_j
        return obj



"""
Custom Datasets
===========================

This example demonstrates how to use your own multiview datasets with CCA-Zoo.
"""

import numpy as np

# %%
# Imports
# -----
import pytorch_lightning as pl

# %% NumpyDataset
# --------------------------------------------------------------
# This is arguably the easiest way to
# use your own data with CCA-Zoo. You can use the NumpyDataset class to convert your data into a valid dataset. This
# class takes in a list of numpy arrays, each representing a view. It also takes in a list of labels, which can be
# None if you don't have labels. The NumpyDataset class will automatically convert your data into a valid dataset.
# You can then use the get_dataloaders function to convert your dataset into dataloaders, which can be used with
# CCA-Zoo.
from cca_zoo.data.deep import NumpyDataset
from cca_zoo.deepmodels import (
    DCCA,
)
from cca_zoo.deepmodels import architectures

X = np.random.normal(size=(100, 10))
Y = np.random.normal(size=(100, 10))
Z = np.random.normal(size=(100, 10))
numpy_dataset = NumpyDataset([X, Y, Z], labels=None)

# %% Checking the dataset
# ------------------------------------------------------
# You can use the check_dataset function to check if your dataset is valid. This function will check if your
# dataset is a valid CCA-Zoo compatible dataset.
from cca_zoo.data.deep import check_dataset

check_dataset(numpy_dataset)

# %% Your own dataset class
# ------------------------------------------------------
# You can also create your own dataset class. This class must inherit from the torch._utils.data.Dataset class.
# It must also have a __getitem__ method, which returns a tuple of views and a label. The views must be a tuple
# of torch tensors.
import torch


class CustomDataset(torch.utils.data.Dataset):
    def __init__(self):
        pass

    def __len__(self):
        return 10

    def __getitem__(self, index):
        return {"views": (torch.rand(10), torch.rand(10))}


custom_dataset = CustomDataset()
check_dataset(custom_dataset)

# %% Using your own dataset
# ------------------------------------------------------
# You can then use the get_dataloaders function to convert your dataset into dataloaders, which can be used with
# CCA-Zoo.
from cca_zoo.data.deep import get_dataloaders

train_loader = get_dataloaders(custom_dataset, batch_size=2)

# %%
# Training
# -----
LATENT_DIMS = 1
EPOCHS = 10

encoder_1 = architectures.Encoder(latent_dims=LATENT_DIMS, feature_size=10)
encoder_2 = architectures.Encoder(latent_dims=LATENT_DIMS, feature_size=10)

# %%
# Deep CCA
# ----------------------------
dcca = DCCA(latent_dims=LATENT_DIMS, encoders=[encoder_1, encoder_2])
trainer = pl.Trainer(
    max_epochs=EPOCHS,
    enable_checkpointing=False,
    log_every_n_steps=1,
)
trainer.fit(dcca, train_loader)



from ._gcca import GCCA, KGCCA
from ._grcca import GRCCA
from ._iterative import (
    PLS_ALS,
    SCCA_PMD,
    ElasticCCA,
    SCCA_Parkhomenko,
    SCCA_IPLS,
    SCCA_ADMM,
    SCCA_Span,
    SWCCA,
    AltMaxVar,
)
from ._mcca import MCCA, KCCA
from ._ncca import NCCA
from ._partialcca import PartialCCA
from ._prcca import PRCCA
from ._rcca import rCCA, CCA, PLS
from ._tcca import TCCA, KTCCA

__all__ = [
    "GCCA",
    "KGCCA",
    "PLS_ALS",
    "SCCA_PMD",
    "ElasticCCA",
    "SCCA_Parkhomenko",
    "SCCA_IPLS",
    "SCCA_ADMM",
    "SCCA_Span",
    "SWCCA",
    "MCCA",
    "KCCA",
    "NCCA",
    "PartialCCA",
    "rCCA",
    "CCA",
    "PLS",
    "TCCA",
    "KTCCA",
    "PRCCA",
    "GRCCA",
    "AltMaxVar",
]

try:
    from ._stochastic import (
        # PLSStochasticPower,
        PLSGHAGEP,
        CCAGHAGEP,
        PLSEigenGame,
        CCAEigenGame,
    )

    __all__.extend(
        # "StochasticPowerPLS",
        "PLSGHAGEP",
        "CCAGHAGEP",
        "PLSEigenGame",
        "CCAEigenGame",
    )
except:
    pass

classes = __all__



import warnings
from typing import Iterable

import numpy as np


def _check_views(views: Iterable[np.ndarray]):
    if len(views) == 2:
        # This is a bit of a hack to try to match up with the way mvlearn takes views which in turn is a bit of a hack to match up with sklearn.
        # Sklearn expects fit(views,y) so if we want multiview views in sklearn functions we need views to be a list
        if isinstance(views[0], list) and views[1] is None:
            views = views[0]

    elif len(views) < 2:
        raise ValueError("Require at least 2 views")

    if not len(set([view.shape[0] for view in views])) == 1:
        msg = "All views must have the same number of samples"
        raise ValueError(msg)


def _process_parameter(parameter_name: str, parameter, default, n_views: int):
    if parameter is None:
        parameter = [default] * n_views
    elif not isinstance(parameter, (list, tuple)):
        parameter = [parameter] * n_views
    _check_parameter_number(parameter_name, parameter, n_views)
    return parameter


def _check_parameter_number(parameter_name: str, parameter, n_views: int):
    if len(parameter) != n_views:
        raise ValueError(
            f"number of views passed should match number of parameter {parameter_name}"
            f"len(views)={n_views} and "
            f"len({parameter_name})={len(parameter)}"
        )


def _check_converged_weights(weights, view_index):
    """check the converged weights are not zero."""
    if np.linalg.norm(weights) <= 0:
        warnings.warn(
            f"All result weights are zero in view {view_index}. "
            "Try less regularisation or another initialisation"
        )


def _check_Parikh2014(mus, lams, views):
    """Return index of the view which the data not matching the condition
    documented in Parikh 2014."""
    failed_check = [
        i
        for i, (mu, lam, view) in enumerate(zip(mus, lams, views))
        if mu < lam / np.linalg.norm(view) ** 2
    ]
    if failed_check:
        raise ValueError(
            "mu, lam, view not matching condition specified "
            "from Parikh 2014 (mu<lam/frobenius(views)**2)."
            "Index of view(s) not meeting the condition: "
            f"{failed_check}."
        )


def _check_batch_size(batch_size, latent_dimensions):
    """check batch size greater than number of latent dimensions and warn user otherwise"""
    if batch_size < latent_dimensions:
        warnings.warn(
            "Objective is unstable when batch size is less than the number of latent dimensions"
        )



import itertools
from abc import abstractmethod
from typing import Union, Iterable

import numpy as np
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_random_state, check_is_fitted, FLOAT_DTYPES

from cca_zoo.utils.check_values import _check_views


class _BaseCCA(BaseEstimator, MultiOutputMixin, RegressorMixin):
    """
    Base class for CCA methods.
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale=True,
        centre=True,
        copy_data=True,
        accept_sparse=False,
        random_state: Union[int, np.random.RandomState] = None,
    ):
        """
        Parameters
        ----------
        latent_dims : int, optional
            Number of latent dimensions to fit. Default is 1.
        scale : bool, optional
            Whether to scale the data to unit variance. Default is True.
        centre : bool, optional
            Whether to centre the data. Default is True.
        copy_data : bool, optional
            Whether to copy the data. Default is True.
        accept_sparse : bool, optional
            Whether to accept sparse data. Default is False.
        random_state : int, RandomState instance or None, optional (default=None)
            Pass an int for reproducible output across multiple function calls.

        """
        self.latent_dims = latent_dims
        self.scale = scale
        self.centre = centre
        self.copy_data = copy_data
        self.accept_sparse = accept_sparse
        self.random_state = check_random_state(random_state)
        self.dtypes = FLOAT_DTYPES

    @abstractmethod
    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        """
        Fits the model to the given data

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs: any additional keyword arguments required by the given model

        Returns
        -------
        self : object

        """
        return self

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        """

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        check_is_fitted(self, attributes=["weights"])
        _check_views(views)
        views = [self.scalers[i].transform(view) for i, view in enumerate(views)]
        transformed_views = []
        for i, view in enumerate(views):
            transformed_view = view @ self.weights[i]
            transformed_views.append(transformed_view)
        return transformed_views

    def fit_transform(self, views: Iterable[np.ndarray], **kwargs):
        """
        Fits the model to the given data and returns the transformed views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        transformed_views : list of numpy arrays

        """
        return self.fit(views, **kwargs).transform(views, **kwargs)

    def get_factor_loadings(
        self, views: Iterable[np.ndarray], normalize=True, **kwargs
    ):
        """
        Returns the factor loadings for each view

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        normalize : bool, optional
            Whether to normalize the factor loadings. Default is True.
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        factor_loadings : list of numpy arrays

        """
        transformed_views = self.transform(views, **kwargs)
        if normalize:
            loadings = [
                np.corrcoef(view, transformed_view, rowvar=False)[
                    : view.shape[1], view.shape[1] :
                ]
                for view, transformed_view in zip(views, transformed_views)
            ]
        else:
            loadings = [
                view.T @ transformed_view
                for view, transformed_view in zip(views, transformed_views)
            ]
        return loadings

    def pairwise_correlations(self, views: Iterable[np.ndarray], **kwargs):
        """
        Returns the pairwise correlations between the views in each dimension

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        pairwise_correlations : numpy array of shape (n_views, n_views, latent_dims)

        """
        transformed_views = self.transform(views, **kwargs)
        all_corrs = []
        for x, y in itertools.product(transformed_views, repeat=2):
            all_corrs.append(
                np.diag(np.corrcoef(x.T, y.T)[: self.latent_dims, self.latent_dims :])
            )
        all_corrs = np.array(all_corrs).reshape(
            (len(transformed_views), len(transformed_views), self.latent_dims)
        )
        return all_corrs

    def score(self, views: Iterable[np.ndarray], y=None, **kwargs):
        """
        Returns the average pairwise correlation between the views

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)
        y : None
        kwargs : any additional keyword arguments required by the given model

        Returns
        -------
        score : float


        """
        # by default return the average pairwise correlation in each dimension (for 2 views just the correlation)
        pair_corrs = self.pairwise_correlations(views, **kwargs)
        # n views
        n_views = pair_corrs.shape[0]
        # sum all the pairwise correlations for each dimension. Subtract the self correlations. Divide by the number of views. Gives average correlation
        dim_corrs = (
            pair_corrs.sum(axis=tuple(range(pair_corrs.ndim - 1))) - n_views
        ) / (n_views**2 - n_views)
        return dim_corrs

    def _check_params(self):
        pass

    def _validate_inputs(self, views):
        """
        Checks that the input data is valid

        Parameters
        ----------
        views : list/tuple of numpy arrays or array likes with the same number of rows (samples)

        Returns
        -------
        views : list of numpy arrays

        """
        # if doesn't have tag multiview=True then check number of views equals 2
        if not self._get_tags().get("multiview", False):
            if len(views) != 2:
                raise ValueError("Only two views are supported for this model.")
        _check_views(views)
        views = [
            self._validate_data(
                view,
                accept_sparse=["csr", "csc", "coo"],
                dtype=self.dtypes,
                copy=self.copy_data,
            )
            for view in views
        ]

        self.scalers = [
            StandardScaler(
                copy=self.copy_data, with_mean=self.centre, with_std=self.scale
            )
            for _ in range(len(views))
        ]
        views = [
            scaler.fit_transform(view) for view, scaler in zip(views, self.scalers)
        ]
        self.n = views[0].shape[0]
        self.n_views = len(views)
        return views



import numpy as np

from cca_zoo.models._stochastic._eigengame import CCAEigenGame, PLSEigenGame


class CCAGHAGEP(CCAEigenGame):
    """
    A class used to fit Regularized CCA by GHA-GEP

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    shuffle : bool, optional
        Whether to shuffle the data, by default True
    sampler : torch.utils.data.Sampler, optional
        Sampler to use, by default None
    batch_sampler : torch.utils.data.Sampler, optional
        Batch sampler to use, by default None
    num_workers : int, optional
        Number of workers to use, by default 0
    pin_memory : bool, optional
        Whether to pin memory, by default False
    drop_last : bool, optional
        Whether to drop the last batch, by default True
    timeout : int, optional
        Timeout to use, by default 0
    worker_init_fn : function, optional
        Worker init function to use, by default None
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01
    c : float, optional
        Regularization parameter, by default 0

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=None,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        learning_rate=1,
        c=0,
        momentum=0,
        line_search=False,
        rho=0.1,
        ensure_descent=False,
        initialization="random",
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
            tol=tol,
            accept_sparse=accept_sparse,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            epochs=epochs,
            learning_rate=learning_rate,
            momentum=momentum,
            line_search=line_search,
            rho=rho,
            ensure_descent=ensure_descent,
            initialization=initialization,
        )

    def grads(self, views, u=None):
        Aw, Bw, wAw, wBw = self._get_terms(views, u)
        return -Aw + Bw @ np.triu(wAw)


class PLSGHAGEP(PLSEigenGame):
    """
    A class used to fit PLS by GHA-GEP

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    shuffle : bool, optional
        Whether to shuffle the data, by default True
    sampler : torch.utils.data.Sampler, optional
        Sampler to use, by default None
    batch_sampler : torch.utils.data.Sampler, optional
        Batch sampler to use, by default None
    num_workers : int, optional
        Number of workers to use, by default 0
    pin_memory : bool, optional
        Whether to pin memory, by default False
    drop_last : bool, optional
        Whether to drop the last batch, by default True
    timeout : int, optional
        Timeout to use, by default 0
    worker_init_fn : function, optional
        Worker init function to use, by default None
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=1,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        learning_rate=1,
        momentum=0,
        line_search=False,
        rho=0.1,
        ensure_descent=True,
        initialization="random",
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            tol=tol,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            epochs=epochs,
            learning_rate=learning_rate,
            momentum=momentum,
            line_search=line_search,
            rho=rho,
            ensure_descent=ensure_descent,
            initialization=initialization,
        )

    def grads(self, views, u=None):
        Aw, Bw, wAw, wBw = self._get_terms(views, u)
        return -Aw + Bw @ np.triu(wAw)



from typing import Union, Iterable

import numpy as np

from cca_zoo.models._iterative._base import _BaseIterative
from cca_zoo.models._search import _delta_search, support_threshold
from cca_zoo.utils import _process_parameter


class SCCA_Span(_BaseIterative):
    r"""
    Fits a Sparse CCA model using SpanCCA.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \text{l1_ratio}\|w_i\|_1\}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

    References
    ----------
    Asteris, Megasthenis, et al. "A simple and provable algorithm for sparse diagonal CCA." International Conference on Machine Learning. PMLR, 2016.


    Examples
    --------
    >>> from cca_zoo.models import SCCA_Span
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_Span(regularisation="l0", tau=[2, 2])
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.84556666])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        max_iter: int = 100,
        initialization: str = "uniform",
        tol: float = 1e-3,
        regularisation="l0",
        tau: Union[Iterable[Union[float, int]], Union[float, int]] = None,
        rank=1,
        positive: Union[Iterable[bool], bool] = None,
        random_state=None,
        deflation="cca",
        verbose=0,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            verbose=verbose,
        )
        self.tau = tau
        self.regularisation = regularisation
        self.rank = rank
        self.positive = positive

    def _check_params(self):
        """check number of views=2"""
        if self.n_views != 2:
            raise ValueError(f"SCCA_Span requires only 2 views")
        self.max_obj = 0
        if self.regularisation == "l0":
            self.update = support_threshold
            self.tau = _process_parameter("tau", self.tau, 0, self.n_views)
        elif self.regularisation == "l1":
            self.update = _delta_search
            self.tau = _process_parameter("tau", self.tau, 0, self.n_views)
        self.positive = _process_parameter(
            "positive", self.positive, False, self.n_views
        )

    def _initialize(self, views):
        self.max_obj = 0
        cov = views[0].T @ views[1] / views[0].shape[0]
        # Perform SVD on im and obtain individual matrices
        P, D, Q = np.linalg.svd(cov, full_matrices=True)
        self.P = P[:, : self.rank]
        self.D = D[: self.rank]
        self.Q = Q[: self.rank, :].T

    def _update(self, views, scores, weights):
        c = self.random_state.randn(self.rank)
        c /= np.linalg.norm(c)
        a = self.P @ np.diag(self.D) @ c
        u = self.update(a, self.tau[0])
        u /= np.linalg.norm(u)
        b = self.Q @ np.diag(self.D) @ self.P.T @ u
        v = self.update(b, self.tau[1])
        v /= np.linalg.norm(v)
        if b.T @ v > self.max_obj:
            self.max_obj = b.T @ v
            scores[0] = views[0] @ u
            scores[1] = views[1] @ v
            weights[0] = u
            weights[1] = v
        return scores, weights



from typing import Union

import numpy as np

from ._base import _BaseStochastic, tv


class IncrementalPLS(_BaseStochastic):
    r"""
    A class used to fit Incremental PLS

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    shuffle : bool, optional
        Whether to shuffle the data, by default True
    sampler : torch.utils.data.Sampler, optional
        Sampler to use, by default None
    batch_sampler : torch.utils.data.Sampler, optional
        Batch sampler to use, by default None
    num_workers : int, optional
        Number of workers to use, by default 0
    pin_memory : bool, optional
        Whether to pin memory, by default False
    drop_last : bool, optional
        Whether to drop the last batch, by default True
    timeout : int, optional
        Timeout to use, by default 0
    worker_init_fn : function, optional
        Worker init function to use, by default None
    epochs : int, optional
        Number of epochs to use, by default 1
    simple : bool, optional
        Whether to use the simple update, by default False

    References
    ----------
    Arora, Raman, et al. "Stochastic optimization for PCA and PLS." 2012 50th Annual Allerton Conference on Communication, Control, and Computing (Allerton). IEEE, 2012.
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        accept_sparse=None,
        batch_size=1,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        simple=False,
        initialization: Union[str, callable] = "random",
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            epochs=epochs,
            initialization=initialization,
        )
        self.simple = simple

    def _update(self, views):
        if not hasattr(self, "S"):
            self.S = np.zeros(self.latent_dims)
            self.count = 0
        if self.simple:
            self.simple_update(views)
        else:
            self.incremental_update(views)
        return False

    def incremental_update(self, views):
        hats = np.stack([view @ weight for view, weight in zip(views, self.weights)])
        orths = [
            view - hat @ weight.T
            for view, weight, hat in zip(views, self.weights, hats)
        ]
        self.incrsvd(hats, orths)

    def simple_update(self, views):
        if not hasattr(self, "M"):
            self.M = np.zeros((views[0].shape[1], views[1].shape[1]))
        self.M = (
            views[0].T @ views[1]
            + self.weights[0] @ np.diag(self.S) @ self.weights[1].T
        )
        U, S, Vt = np.linalg.svd(self.M)
        self.weights[0] = U[:, : self.latent_dims]
        self.weights[1] = Vt.T[:, : self.latent_dims]
        self.S = S[: self.latent_dims]

    def incrsvd(self, hats, orths):
        Q = np.vstack(
            (
                np.hstack(
                    (
                        np.diag(self.S) + hats[0].T @ hats[1],
                        np.linalg.norm(orths[1], axis=1).T * hats[0].T,
                    )
                ),
                np.hstack(
                    (
                        (np.linalg.norm(orths[0], axis=1).T * hats[1].T).T,
                        np.atleast_2d(
                            np.linalg.norm(orths[0], axis=1, keepdims=True)
                            @ np.linalg.norm(orths[1], axis=1, keepdims=True).T
                        ),
                    )
                ),
            )
        )
        U, S, Vt = np.linalg.svd(Q)
        self.weights[0] = (
            np.hstack((self.weights[0], orths[0].T / np.linalg.norm(orths[0])))
            @ U[:, : self.latent_dims]
        )
        self.weights[1] = (
            np.hstack((self.weights[1], orths[1].T / np.linalg.norm(orths[1])))
            @ Vt.T[:, : self.latent_dims]
        )
        self.S = S[: self.latent_dims]

    def objective(self, views, **kwargs):
        q = [np.linalg.qr(weight)[0] for weight in self.weights]
        views = [self.scalers[i].transform(view) for i, view in enumerate(views)]
        transformed_views = []
        for i, (view) in enumerate(views):
            transformed_view = view @ q[i]
            transformed_views.append(transformed_view)
        return tv(transformed_views)



from abc import abstractmethod
from typing import Iterable, Union

import numpy as np
from scipy.linalg import block_diag
from torch.utils import data
from torch.utils.data import BatchSampler, DataLoader, RandomSampler, SequentialSampler

from cca_zoo.data.deep import NumpyDataset
from cca_zoo.models import CCA
from cca_zoo.models._base import _BaseCCA
from cca_zoo.models._iterative._base import _default_initializer


class _BaseStochastic(_BaseCCA):
    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        tol=1e-3,
        accept_sparse=None,
        batch_size=None,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        val_split=None,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        track_training=False,
        momentum=0,
    ):
        if accept_sparse is None:
            accept_sparse = ["csc", "csr"]
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
        )
        self.tol = tol
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.sampler = sampler
        self.batch_sampler = batch_sampler
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.drop_last = drop_last
        self.timeout = timeout
        self.worker_init_fn = worker_init_fn
        self.epochs = epochs
        self.val_split = val_split
        self.learning_rate = learning_rate
        self.initialization = initialization
        self.track_training = track_training
        self.momentum = momentum
        self.dtypes = [np.float32]

    def fit(self, views: Iterable[np.ndarray], y=None, **kwargs):
        views = self._validate_inputs(views)
        self._check_params()
        train_dataloader, val_dataloader = self.get_dataloader(views)
        self.track = []
        initializer = _default_initializer(
            views, self.initialization, self.random_state, self.latent_dims
        )
        self.weights = initializer.fit(views).weights
        self.weights = [weights.astype(np.float32) for weights in self.weights]
        stop = False
        for _ in range(self.epochs):
            for i, sample in enumerate(train_dataloader):
                stop = self._update(sample["views"])
            if self.val_split is not None:
                for i, sample in enumerate(val_dataloader):
                    self.track.append(self.objective(sample["views"], u=self.weights))
            else:
                self.track.append(self.objective(views, u=self.weights))
            if stop:
                break
        return self

    def get_dataloader(self, views: Iterable[np.ndarray]):
        if self.batch_size is None:
            dataset = BatchNumpyDataset(views)
        else:
            dataset = NumpyDataset(views)
        if self.val_split is not None:
            train_size = int((1 - self.val_split) * len(dataset))
            val_size = len(dataset) - train_size
            dataset, val_dataset = data.random_split(dataset, [train_size, val_size])
            sampler = BatchSampler(
                SequentialSampler(dataset), batch_size=len(val_dataset), drop_last=False
            )
            val_loader = DataLoader(
                val_dataset,
                sampler=sampler,
                num_workers=self.num_workers,
                pin_memory=self.pin_memory,
                drop_last=self.drop_last,
                timeout=self.timeout,
                worker_init_fn=self.worker_init_fn,
                collate_fn=lambda x: x[0],
            )
        else:
            val_loader = None
        if self.batch_size is None:
            batch_size = len(dataset)
        else:
            batch_size = self.batch_size
        sampler = BatchSampler(
            RandomSampler(dataset), batch_size=batch_size, drop_last=False
        )
        train_loader = DataLoader(
            dataset,
            sampler=sampler,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            timeout=self.timeout,
            worker_init_fn=self.worker_init_fn,
            collate_fn=lambda x: x[0],
        )
        return train_loader, val_loader

    @abstractmethod
    def _update(self, views):
        pass

    @abstractmethod
    def objective(self, views, weights):
        raise NotImplementedError


class BatchNumpyDataset(NumpyDataset):
    def __getitem__(self, index):
        if self.labels is not None:
            return {"views": self.views, "label": self.labels}
        else:
            return {"views": self.views}


def tv(z):
    all_z = np.hstack(z)
    C = np.cov(all_z, rowvar=False)
    C -= block_diag(*[np.cov(z_, rowvar=False) for z_ in z])
    C /= z[0].shape[0]
    return np.linalg.svd(C, compute_uv=False).sum()


def tcc(z):
    return CCA(z[0].shape[1]).fit(z).score(z).sum()



from ._eigengame import PLSEigenGame, CCAEigenGame
from ._ghagep import PLSGHAGEP, CCAGHAGEP

__all__ = [
    # "PLSStochasticPower",
    "PLSGHAGEP",
    "CCAGHAGEP",
    "PLSEigenGame",
    "CCAEigenGame",
]



from typing import Iterable, Union

import numpy as np
import numpy.linalg as la
from sklearn.kernel_approximation import Nystroem
from sklearn.metrics import pairwise_kernels
from sklearn.model_selection import train_test_split
from sklearn.utils.validation import check_is_fitted

from cca_zoo.models._iterative._base import _BaseIterative
from cca_zoo.utils import _process_parameter, _check_views


class GradKCCA(_BaseIterative):
    """
    References
    ----------
    [1] Viivi Uurtio, Sahely Bhadra, and Juho Rousu. Large-scale sparse kernel canonical correlation analysis. In Kamalika Chaudhuri and Ruslan Salakhutdinov, editors, Proceedings of the 36th International Conference on Machine Learning, volume 97 of Proceedings of Machine Learning Research, pages 6383–6391, Long Beach, California, USA, 09–15 Jun 2019. PMLR.

    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        proj: Union[Iterable[float], float] = "l1",
        kernel: Iterable[Union[float, callable]] = None,
        gamma: Iterable[float] = None,
        degree: Iterable[float] = None,
        coef0: Iterable[float] = None,
        kernel_params: Iterable[dict] = None,
        repetitions=5,
        initialization: Union[str, callable] = "random",
        nystrom=False,
        nystrom_components=100,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
            initialization=initialization,
        )
        self.proj = proj
        self.kernel_params = kernel_params
        self.gamma = gamma
        self.coef0 = coef0
        self.kernel = kernel
        self.degree = degree
        self.repetitions = repetitions
        self.nystrom = nystrom
        self.nystrom_components = nystrom_components

    def _check_params(self):
        self.proj = _process_parameter("proj", self.proj, "l1", self.n_views)
        self.kernel = _process_parameter("kernel", self.kernel, "linear", self.n_views)
        self.gamma = _process_parameter("gamma", self.gamma, None, self.n_views)
        self.coef0 = _process_parameter("coef0", self.coef0, 1, self.n_views)
        self.degree = _process_parameter("degree", self.degree, 1, self.n_views)
        self.nystrom_components = _process_parameter(
            "nystrom components", self.nystrom_components, 100, self.n_views
        )

    def _get_kernel(self, view_index, X, Y=None):
        if callable(self.kernel[view_index]):
            params = self.kernel_params[view_index] or {}
        else:
            params = {
                "gamma": self.gamma[view_index],
                "degree": self.degree[view_index],
                "coef0": self.coef0[view_index],
            }
        if self.nystrom:
            nystroem = Nystroem(
                kernel=self.kernel[view_index],
                n_components=self.nystrom_components[view_index],
                **params,
            )
            if Y is None:
                Y = X
            K = nystroem.fit_transform(X) @ nystroem.transform(Y).T
            return K
        else:
            return pairwise_kernels(
                X, Y, metric=self.kernel[view_index], filter_params=True, **params
            )

    def backracking_line_search(self, w, gw, stp, X, K, obj_old, view_index):
        while True:
            w_new = w + gw * stp
            w_new = w_new / la.norm(w_new)
            Kw_new = self._get_kernel(view_index, X, w_new[None, :])
            obj_new = self._objective(None, (Kw_new, K), None)
            if obj_new > obj_old + 1e-4 * np.abs(obj_old):
                w = w_new
                break
            elif stp < 1e-7:
                w = w_new
                break
            else:
                stp /= 2
        return w

    def _objective(self, views, scores, weights) -> int:
        return (
            scores[0].T
            @ scores[1]
            / (np.sqrt(scores[0].T @ scores[0]) * np.sqrt(scores[1].T @ scores[1]))
        )

    def _proj_l1(self, v, b):
        assert b > 0
        if la.norm(v, 1) < b:
            return v
        u = -np.sort(-np.abs(v))
        sv = np.cumsum(u)
        r = np.where(u > (sv - b) / np.arange(1, u.shape[0] + 1))
        if len(r[-1]) > 0:
            rho = r[-1][-1]
            tau = (sv[rho] - b) / rho
            theta = np.maximum(0, tau)
            return np.sign(v) * np.maximum(np.abs(v) - theta, 0)
        else:
            return v

    def _proj_l2(self, v, b):
        return v / la.norm(v) / b

    def _update(self, views, scores, weights):
        scores = [
            self._get_kernel(i, view, weights[i][None, :])
            for i, view in enumerate(views)
        ]
        obj_old = self._objective(views, scores, weights)
        for view_index, view in enumerate(views):
            # TODO gradients
            grad = 0
            gamma = la.norm(grad)
            weights[view_index] = self.backracking_line_search(
                weights[view_index],
                grad,
                gamma,
                view,
                scores[view_index],
                obj_old,
                view_index,
            )
            if self.proj[view_index] == "l1":
                weights[view_index] = self._proj_l1(
                    weights[view_index], self.c[view_index]
                )
            elif self.proj[view_index] == "l2":
                weights[view_index] = self._proj_l2(
                    weights[view_index], self.c[view_index]
                )
            else:
                raise ValueError(
                    "projection {self.proj[view_index]} not supported. Pass a generator implementing this method"
                )
        return scores, weights

    def transform(self, views: Iterable[np.ndarray], **kwargs):
        check_is_fitted(self, attributes=["weights"])
        views = _check_views(*views)
        views = self._centre_scale_transform(views)
        scores = [
            self._get_kernel(i, view, self.weights[i].T) for i, view in enumerate(views)
        ]
        return scores


def generate_data(n, p, q):
    X = np.random.uniform(-1, 1, [n, p])
    Y = np.random.uniform(-1, 1, [n, q])
    Y[:, 2] = X[:, 2] + X[:, 3] - Y[:, 3] + np.random.normal(0, 0.05, n)
    # Y[:,2] = np.power(X[:,2] + X[:,3],3) - Y[:,3] + np.random.normal(0,0.05,n)
    # Y[:,4] = np.exp(X[:,4] + X[:,5]) - Y[:,5] + np.random.normal(0,0.05,n)
    return X, Y


def main():
    import matplotlib.pyplot as plt

    np.set_printoptions(precision=2)

    X, Y = generate_data(1000, 8, 8)
    Xtrain, Xtest, Ytrain, Ytest = train_test_split(X, Y)

    kernel1 = GradKCCA().fit((Xtrain, Ytrain))
    print(f"Training Correlation: {kernel1.score((Xtrain, Ytrain))}")
    print(f"Test Correlation: {kernel1.score((Xtest, Ytest))}")

    plt.plot(*kernel1.transform((Xtrain, Ytrain)), "bo")
    plt.show()


if __name__ == "__main__":
    main()



import torch
from ._dcca_eigengame import DCCA_EigenGame


class DCCA_GHA(DCCA_EigenGame):
    def loss(self, views, **kwargs):
        z = self(views)
        A, B = self.get_AB(z)
        rewards = torch.trace(A) + torch.trace(A).detach()
        penalties = torch.trace(A.detach() @ B)
        return {
            "objective": -rewards.sum() + penalties,
            "rewards": rewards.sum(),
            "penalties": penalties,
        }



import torch
import torchvision
from pytorch_lightning import Callback
from pytorch_lightning import LightningModule, Trainer
from torch.autograd import Variable


class CorrelationCallback(Callback):
    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        pl_module.log(
            "val/corr",
            pl_module.score(trainer.val_dataloaders[0]).sum(),
        )


class GenerativeCallback(Callback):
    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ) -> None:
        if hasattr(pl_module, "img_dim") and pl_module.img_dim is not None:
            z = dict()
            z["shared"] = Variable(torch.randn(64, pl_module.latent_dims))
            if pl_module.private_encoders:
                z["private"] = [Variable(torch.randn(64, pl_module.latent_dims))] * len(
                    pl_module.private_encoders
                )
            sample = pl_module._decode(z)
            sample[0] = torch.reshape(sample[0], (64,) + pl_module.img_dim)
            sample[1] = torch.reshape(sample[1], (64,) + pl_module.img_dim)
            grid1 = torchvision.utils.make_grid(sample[0])
            grid2 = torchvision.utils.make_grid(sample[1])
            pl_module.logger.experiment.add_image(
                "generated_images_1", grid1, pl_module.current_epoch
            )
            pl_module.logger.experiment.add_image(
                "generated_images_2", grid2, pl_module.current_epoch
            )



from typing import Union

import numpy as np

from cca_zoo.models._stochastic._base import _BaseStochastic


class CCAEigenGame(_BaseStochastic):
    """
    A class used to fit Regularized CCA by Delta-EigenGame

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    shuffle : bool, optional
        Whether to shuffle the data, by default True
    sampler : torch.utils.data.Sampler, optional
        Sampler to use, by default None
    batch_sampler : torch.utils.data.Sampler, optional
        Batch sampler to use, by default None
    num_workers : int, optional
        Number of workers to use, by default 0
    pin_memory : bool, optional
        Whether to pin memory, by default False
    drop_last : bool, optional
        Whether to drop the last batch, by default True
    timeout : int, optional
        Timeout to use, by default 0
    worker_init_fn : function, optional
        Worker init function to use, by default None
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01
    c : float, optional
        Regularization parameter, by default 0

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=None,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        learning_rate=1,
        initialization: Union[str, callable] = "random",
        momentum=0,
        line_search=False,
        rho=0.1,
        ensure_descent=False,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            tol=tol,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            epochs=epochs,
            learning_rate=learning_rate,
            initialization=initialization,
            momentum=momentum,
        )
        self.rho = rho
        self.line_search = line_search
        self.ensure_descent = ensure_descent
        self.velocity = None

    def _split_weights(self, views, weights):
        splits = np.cumsum([0] + [view.shape[1] for view in views])
        weights = [
            weights[split : splits[i + 1]] for i, split in enumerate(splits[:-1])
        ]
        return weights

    def _update(self, views):
        # combine weights
        self.weights = np.vstack(self.weights)
        self._update_all(views)
        # split weights
        self.weights = self._split_weights(views, self.weights)

    def _update_all(self, views):
        if self.velocity is None:
            self.velocity = np.zeros_like(self.weights)
        self.weights += self.momentum * self.velocity
        step_direction = self.grads(views, u=self.weights)
        if self.line_search:
            self.learning_rate = self._backtracking_line_search(views, step_direction)
        step_size = self.learning_rate
        self.velocity = self.momentum * self.velocity - step_size * step_direction
        self.weights = self._gradient_step(self.weights, self.velocity)

    def _backtracking_line_search(self, views, step_direction, beta=0.9):
        # Initialize the gradient vector
        gradient = self.grads(views, u=self.weights)

        # Initialize the step size
        if self.momentum == 0:
            step_size = self.learning_rate
        else:
            step_size = 1

        # Perform backtracking line search until the objective function decreases
        while True:
            # Update the solution vector based on the gradient and step size
            updated_solution = self._gradient_step(
                self.weights, -step_direction * step_size
            )

            # Check if the objective function decreases with the updated solution
            if self.objective(views, u=updated_solution) < self.objective(
                views, u=self.weights
            ) - 0.1 * step_size * np.linalg.norm(step_direction.T @ gradient):
                break

            # Reduce the step size by a factor of beta
            step_size *= beta

            # Set a lower bound for the step size to avoid numerical issues
            if step_size < 1e-3:
                step_size = 1e-3
                break

        return step_size

    def grads(self, views, u=None):
        Aw, Bw, wAw, wBw = self._get_terms(views, u)
        grads = 2 * Aw - (Aw @ np.triu(wBw) * np.sign(np.diag(wAw)) + Bw @ np.triu(wAw))
        return -grads

    def _Aw(self, views, projections):
        Aw = np.vstack(
            [
                np.cov(view.T, projections.sum(axis=0).T)[
                    0 : view.shape[1], view.shape[1] :
                ]
                for view in views
            ]
        )
        return Aw / len(views)

    def _Bw(self, views, projections, u):
        weights = self._split_weights(views, u)
        Bw = []
        for i, (view, projection, weight) in enumerate(
            zip(views, projections, weights)
        ):
            Bw.append((view.T @ projection) / (projection.shape[0] - 1))
        return np.vstack(Bw) / len(views)

    def _get_terms(self, views, u, projections=None):
        if projections is None:
            projections = self.projections(views, u)
        Aw = self._Aw(views, projections)
        Bw = self._Bw(views, projections, u)
        wAw = u.T @ Aw
        wBw = u.T @ Bw
        return Aw, Bw, wAw, wBw

    def objective(self, views, u=None, k=None, projections=None):
        # if u is a list type
        if isinstance(u, list):
            u = np.vstack(u)
        if k is None:
            k = self.latent_dims
        if projections is None:
            projections = self.projections(views, u)
        Aw, Bw, wAw, wBw = self._get_terms(views, u, projections=projections)
        return -2 * np.trace(wAw[: k + 1, : k + 1]) + np.trace(
            wAw[: k + 1, : k + 1] @ wBw[: k + 1, : k + 1]
        )

    def projections(self, views, u):
        u = self._split_weights(views, u)
        m = []
        for view, weight in zip(views, u):
            n = view @ weight
            m.append(n)
        return np.stack(m)

    def _gradient_step(self, weights, velocity):
        return weights + velocity

    def _more_tags(self):
        return {"multiview": True, "stochastic": True}


class PLSEigenGame(CCAEigenGame):
    """
    A class used to fit PLS by Delta-EigenGame

    Parameters
    ----------
    latent_dims : int, optional
        Number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        Random state to use, by default None
    accept_sparse : bool, optional
        Whether to accept sparse data, by default None
    batch_size : int, optional
        Batch size to use, by default 1
    shuffle : bool, optional
        Whether to shuffle the data, by default True
    sampler : torch.utils.data.Sampler, optional
        Sampler to use, by default None
    batch_sampler : torch.utils.data.Sampler, optional
        Batch sampler to use, by default None
    num_workers : int, optional
        Number of workers to use, by default 0
    pin_memory : bool, optional
        Whether to pin memory, by default False
    drop_last : bool, optional
        Whether to drop the last batch, by default True
    timeout : int, optional
        Timeout to use, by default 0
    worker_init_fn : function, optional
        Worker init function to use, by default None
    epochs : int, optional
        Number of epochs to use, by default 1
    learning_rate : float, optional
        Learning rate to use, by default 0.01

    References
    ----------
    Chapman, James, Ana Lawry Aguila, and Lennie Wells. "A Generalized EigenGame with Extensions to Multiview Representation Learning." arXiv preprint arXiv:2211.11323 (2022).
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        tol=1e-9,
        accept_sparse=None,
        batch_size=None,
        shuffle=True,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        drop_last=True,
        timeout=0,
        worker_init_fn=None,
        epochs=1,
        learning_rate=1,
        momentum=0,
        line_search=False,
        rho=0.1,
        ensure_descent=False,
        initialization="random",
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            accept_sparse=accept_sparse,
            random_state=random_state,
            tol=tol,
            batch_size=batch_size,
            shuffle=shuffle,
            sampler=sampler,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            epochs=epochs,
            learning_rate=learning_rate,
            momentum=momentum,
            line_search=line_search,
            rho=rho,
            ensure_descent=ensure_descent,
            initialization=initialization,
        )

    def _Aw(self, views, projections):
        Aw = np.vstack(
            [
                np.cov(view, projections.sum(axis=0), rowvar=False)[
                    0 : view.shape[1], view.shape[1] :
                ]
                for view in views
            ]
        ) - np.vstack(
            [
                np.cov(view, projection, rowvar=False)[
                    0 : view.shape[1], view.shape[1] :
                ]
                for view, projection in zip(views, projections)
            ]
        )
        return Aw / len(views)

    def _Bw(self, views, projections, u):
        weights = self._split_weights(views, u)
        Bw = []
        for i, (view, projection, weight) in enumerate(
            zip(views, projections, weights)
        ):
            Bw.append(weight)
        return np.vstack(Bw) / len(views)



from typing import Union, Iterable

import numpy as np

from ._base import _BaseIterative
from ...utils import _process_parameter, _check_converged_weights, _check_Parikh2014


class SCCA_ADMM(_BaseIterative):
    r"""
    Fits a sparse CCA model by alternating ADMM for two or more views.

    .. math::

        w_{opt}=\underset{w}{\mathrm{argmax}}\{\sum_i\sum_{j\neq i} \|X_iw_i-X_jw_j\|^2 + \|w_i\|_1\}\\

        \text{subject to:}

        w_i^TX_i^TX_iw_i=1

    Parameters
    ----------
    latent_dims : int, default=1
        Number of latent dimensions to use in the model.
    scale : bool, default=True
        Whether to scale the data to unit variance.
    centre : bool, default=True
        Whether to centre the data to have zero mean.
    copy_data : bool, default=True
        Whether to copy the data or overwrite it.
    random_state : int, default=None
        Random seed for initialisation.
    deflation : str, default="cca"
        Deflation method to use. Options are "cca" and "pls".
    tau : float or list of floats, default=None
        Regularisation parameter. If a single float is given, the same value is used for all views.
        If a list of floats is given, the values are used for each view.
    mu : float or list of floats, default=None
        Regularisation parameter. If a single float is given, the same value is used for all views.
        If a list of floats is given, the values are used for each view.
    lam : float or list of floats, default=None
        Regularisation parameter. If a single float is given, the same value is used for all views.
        If a list of floats is given, the values are used for each view.
    eta : float or list of floats, default=None
        Regularisation parameter. If a single float is given, the same value is used for all views.
        If a list of floats is given, the values are used for each view.
    max_iter : int, default=100
        Maximum number of iterations to run.
    initialization : str or callable, default="pls"
        Method to use for initialisation. Options are "pls" and "random".
    tol : float, default=1e-9
        Tolerance for convergence.
    verbose : int, default=0
        Verbosity level. If 0, no output is printed. If 1, output is printed every 10 iterations.



    References
    ----------
    Suo, Xiaotong, et al. "Sparse canonical correlation analysis." arXiv preprint arXiv:1705.10865 (2017).

    Examples
    --------
    >>> from cca_zoo.models import SCCA_ADMM
    >>> import numpy as np
    >>> rng=np.random.RandomState(0)
    >>> X1 = rng.random((10,5))
    >>> X2 = rng.random((10,5))
    >>> model = SCCA_ADMM(random_state=0,tau=[1e-1,1e-1])
    >>> model.fit((X1,X2)).score((X1,X2))
    array([0.84348183])
    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        deflation="cca",
        tau: Union[Iterable[float], float] = None,
        mu: Union[Iterable[float], float] = None,
        lam: Union[Iterable[float], float] = None,
        eta: Union[Iterable[float], float] = None,
        max_iter: int = 100,
        initialization: Union[str, callable] = "pls",
        tol: float = 1e-3,
        verbose=0,
    ):
        self.tau = tau
        self.mu = mu
        self.lam = lam
        self.eta = eta
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            max_iter=max_iter,
            initialization=initialization,
            tol=tol,
            random_state=random_state,
            deflation=deflation,
            verbose=verbose,
        )

    def _check_params(self):
        self.tau = _process_parameter("tau", self.tau, 0, self.n_views)
        self.lam = _process_parameter("lam", self.lam, 1, self.n_views)
        self.eta = _process_parameter("eta", self.eta, 0, self.n_views)

    def _initialize(self, views):
        self.eta = [np.ones(view.shape[0]) * eta for view, eta in zip(views, self.eta)]
        self.z = [np.zeros(view.shape[0]) for view in views]
        if self.mu is None:
            self.mu = [
                lam / np.linalg.norm(view) ** 2 for lam, view in zip(self.lam, views)
            ]
        else:
            self.mu = _process_parameter("mu", self.mu, 0, self.n_views)
        _check_Parikh2014(self.mu, self.lam, views)

    def _update(self, views, scores, weights):
        # Update each view using loop update function
        for view_index, view in enumerate(views):
            targets = np.ma.array(scores, mask=False)
            targets.mask[view_index] = True
            gradient = views[view_index].T @ targets.sum(axis=0).filled()
            mu = self.mu[view_index]
            lam = self.lam[view_index]
            N = views[view_index].shape[0]
            unnorm_z = []
            norm_eta = []
            norm_weights = []
            norm_proj = []
            for _ in range(self.max_iter):
                # We multiply 'tau' by N in order to make regularisation match across the different sparse cca methods
                weights[view_index] = self._prox_mu_f(
                    weights[view_index]
                    - mu
                    / lam
                    * views[view_index].T
                    @ (
                        views[view_index] @ weights[view_index]
                        - self.z[view_index]
                        + self.eta[view_index]
                    ),
                    mu,
                    gradient,
                    N * self.tau[view_index],
                )
                unnorm_z.append(
                    np.linalg.norm(
                        views[view_index] @ weights[view_index] + self.eta[view_index]
                    )
                )
                self.z[view_index] = self._prox_lam_g(
                    views[view_index] @ weights[view_index] + self.eta[view_index]
                )
                self.eta[view_index] = (
                    self.eta[view_index]
                    + views[view_index] @ weights[view_index]
                    - self.z[view_index]
                )
                norm_eta.append(np.linalg.norm(self.eta[view_index]))
                norm_proj.append(
                    np.linalg.norm(views[view_index] @ weights[view_index])
                )
                norm_weights.append(np.linalg.norm(weights[view_index], 1))
            _check_converged_weights(weights[view_index], view_index)
            scores[view_index] = views[view_index] @ weights[view_index]
        return scores, weights

    def _prox_mu_f(self, x, mu, c, tau):
        u_update = x.copy()
        mask_1 = x + (mu * c) > mu * tau
        # if mask_1.sum()>0:
        u_update[mask_1] = x[mask_1] + mu * (c[mask_1] - tau)
        mask_2 = x + (mu * c) < -mu * tau
        # if mask_2.sum() > 0:
        u_update[mask_2] = x[mask_2] + mu * (c[mask_2] + tau)
        mask_3 = ~(mask_1 | mask_2)
        u_update[mask_3] = 0
        return u_update

    def _prox_lam_g(self, x):
        norm = np.linalg.norm(x)
        if norm < 1:
            return x
        else:
            return x / max(1, norm)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    REQUIRED_PACKAGES = f.read()

EXTRA_PACKAGES = {
    "deep": ["torch", "torchvision", "pytorch-lightning"],
    "probabilistic": ["jax", "numpyro", "arviz"],
}
EXTRA_PACKAGES["all"] = EXTRA_PACKAGES["deep"] + EXTRA_PACKAGES["probabilistic"]

setup(
    name="cca_zoo",
    version="1.17.4",
    include_package_data=True,
    keywords="cca",
    packages=find_packages(),
    url="https://github.com/jameschapman19/cca_zoo",
    license="MIT",
    author="jameschapman",
    description=(
        "Canonical Correlation Analysis Zoo: A collection of Regularized, Deep Learning based, Kernel, and Probabilistic methods in a scikit-learn style framework"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="james.chapman.19@ucl.ac.uk",
    python_requires=">=3.8",
    install_requires=REQUIRED_PACKAGES,
    extras_require=EXTRA_PACKAGES,
    test_suite="test",
    tests_require=[],
)

with open("requirements.txt", "r") as f:
    REQUIRED_PACKAGES = f.read()



import torch
import torch.nn.functional as F

from ._dcca_noi import DCCA_NOI


class DCCA_SDL(DCCA_NOI):
    """
    A class used to fit a Deep CCA by Stochastic Decorrelation model.

    References
    ----------
    Chang, Xiaobin, Tao Xiang, and Timothy M. Hospedales. "Scalable and effective deep CCA via soft decorrelation." Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition. 2018.

    """

    def __init__(
        self,
        latent_dims: int,
        N: int,
        encoders=None,
        r: float = 0,
        rho: float = 0.2,
        eps: float = 1e-5,
        shared_target: bool = False,
        lam=0.5,
        **kwargs
    ):
        super().__init__(
            latent_dims=latent_dims,
            N=N,
            encoders=encoders,
            r=r,
            rho=rho,
            eps=eps,
            shared_target=shared_target,
            **kwargs
        )
        self.c = None
        self.cross_cov = None
        self.lam = lam
        self.bns = torch.nn.ModuleList(
            [torch.nn.BatchNorm1d(latent_dims, affine=False) for _ in self.encoders]
        )

    def forward(self, views, **kwargs):
        z = []
        for i, (encoder, bn) in enumerate(zip(self.encoders, self.bns)):
            z.append(bn(encoder(views[i])))
        return z

    def loss(self, views, **kwargs):
        z = self(views)
        l2_loss = F.mse_loss(z[0], z[1])
        self._update_covariances(z, train=self.training)
        SDL_loss = self._sdl_loss(self.covs)
        loss = l2_loss + self.lam * SDL_loss
        self.covs = [cov.detach() for cov in self.covs]
        return {"objective": loss, "l2": l2_loss, "sdl": SDL_loss}

    def _sdl_loss(self, covs):
        loss = 0
        for cov in covs:
            sgn = torch.sign(cov)
            sgn.fill_diagonal_(0)
            loss += torch.mean(cov * sgn)
        return loss

    def _update_covariances(self, z, train=True):
        b = z[0].shape[0]
        batch_covs = [self.N * z_.T @ z_ / b for z_ in z]
        if train:
            if self.covs is not None:
                self.covs = [
                    self.rho * self.covs[i] + (1 - self.rho) * batch_cov
                    for i, batch_cov in enumerate(batch_covs)
                ]
            else:
                self.covs = batch_covs
        # pytorch-lightning runs validation once so this just fixes the bug
        elif self.covs is None:
            self.covs = batch_covs



from typing import Iterable, Union

import numpy as np
from sklearn.utils.validation import check_is_fitted

from cca_zoo.models import MCCA
from cca_zoo.utils import _check_views


class PartialCCA(MCCA):
    r"""
    A class used to fit a partial cca model. The key difference between this and a vanilla CCA or MCCA is that
    the canonical score vectors must be orthogonal to the supplied confounding variables.

    Parameters
    ----------
    latent_dims : int, optional
        The number of latent dimensions to use, by default 1
    scale : bool, optional
        Whether to scale the data, by default True
    centre : bool, optional
        Whether to centre the data, by default True
    copy_data : bool, optional
        Whether to copy the data, by default True
    random_state : int, optional
        The random state to use, by default None
    c : Union[Iterable[float], float], optional
        The regularization parameter, by default None

    References
    ----------
    Rao, B. Raja. "Partial canonical correlations." Trabajos de estadistica y de investigación operativa 20.2-3 (1969): 211-219.

    Example
    -------
    >>> from cca_zoo.models import PartialCCA
    >>> X1 = np.random.rand(10,5)
    >>> X2 = np.random.rand(10,5)
    >>> partials = np.random.rand(10,3)
    >>> model = PartialCCA()
    >>> model.fit((X1,X2),partials=partials).score((X1,X2))
    array([0.99993046])

    """

    def __init__(
        self,
        latent_dims: int = 1,
        scale: bool = True,
        centre=True,
        copy_data=True,
        random_state=None,
        c: Union[Iterable[float], float] = None,
    ):
        super().__init__(
            latent_dims=latent_dims,
            scale=scale,
            centre=centre,
            copy_data=copy_data,
            random_state=random_state,
        )
        self.c = c

    def fit(self, views: Iterable[np.ndarray], y=None, partials=None, **kwargs):
        super().fit(views, y=y, partials=partials, **kwargs)

    def _setup_evp(self, views: Iterable[np.ndarray], partials=None, **kwargs):
        if partials is None:
            raise ValueError(
                f"partials is {partials}. Require matching partials to transform with"
                f"partial CCA."
            )
        self.confound_betas = [np.linalg.pinv(partials) @ view for view in views]
        views = [
            view - partials @ np.linalg.pinv(partials) @ view
            for view, confound_beta in zip(views, self.confound_betas)
        ]
        return super()._setup_evp(views)

    # TODO TRANSFORM
    def transform(self, views: Iterable[np.ndarray], partials=None, **kwargs):
        if partials is None:
            raise ValueError(
                f"partials is {partials}. Require matching partials to transform with"
                f"partial CCA."
            )
        check_is_fitted(self, attributes=["weights"])
        _check_views(views)
        views = [self.scalers[i].transform(view) for i, view in enumerate(views)]
        transformed_views = []
        for i, (view) in enumerate(views):
            transformed_view = (
                view - partials @ self.confound_betas[i]
            ) @ self.weights[i]
            transformed_views.append(transformed_view)
        return transformed_views

    def _more_tags(self):
        return {"multiview": True}



"""
Model Validation
===========================

This script will show how to use the model validation methods in CCA-Zoo including
permutation testing, learning curves, and cross-validation.
"""

import matplotlib.pyplot as plt
import numpy as np

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.models import CCA

# %%
# Data
# ------
np.random.seed(42)
n = 200
p = 15
q = 15
latent_dims = 3

(X, Y) = LinearSimulatedData(
    view_features=[p, q], latent_dims=latent_dims, correlation=[0.9, 0.5, 0.1]
).sample(n)

# %%
# Permutation Testing
# -------------------

from cca_zoo.model_selection import permutation_test_score, learning_curve
from sklearn.model_selection import KFold, ShuffleSplit

LATENT_DIMS = 3
model = CCA(latent_dims=LATENT_DIMS)
cv = KFold(2, shuffle=True, random_state=0)
score, perm_scores, pvalue = permutation_test_score(
    model, (X, Y), cv=cv, n_permutations=100
)

# %%
fig, ax = plt.subplots(LATENT_DIMS, figsize=[12, 8])
for k in range(LATENT_DIMS):
    ax[k].hist(perm_scores[k])
    ax[k].axvline(score[k], ls="--", color="r")
    score_label = f"Score on original\ndata: {score[k]:.2f}\n(p-value: {pvalue[k]:.3f})"
    ax[k].text(0, 0.5, score_label, fontsize=12)
    ax[k].set_xlabel("Correlation")
    _ = ax[k].set_ylabel("Probability")
    ax[k].set_title(f"Dimension {k + 1}")
plt.tight_layout()
plt.show()

# %%
# Learning Curves
# -------------------

import numpy as np
import matplotlib.pyplot as plt


def plot_learning_curve(
    estimator,
    title,
    views,
    axes=None,
    ylim=None,
    cv=None,
    n_jobs=None,
    train_sizes=np.linspace(0.1, 1.0, 5),
):
    """
    Generate 3 plots: the test and training learning curve, the training
    samples vs fit times curve, the fit times vs score curve.

    Parameters
    ----------
    estimator : estimator instance
        An estimator instance implementing `fit` and `predict` methods which
        will be cloned for each validation.

    title : str
        Title for the chart.

    X : array-like of shape (n_samples, n_features)
        Training vector, where ``n_samples`` is the number of samples and
        ``n_features`` is the number of features.

    y : array-like of shape (n_samples) or (n_samples, n_features)
        Target relative to ``views`` for classification or regression;
        None for unsupervised learning.

    axes : array-like of shape (3,), default=None
        Axes to use for plotting the curves.

    ylim : tuple of shape (2,), default=None
        Defines minimum and maximum y-values plotted, e.g. (ymin, ymax).

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

          - None, to use the default 5-fold cross-validation,
          - integer, to specify the number of folds.
          - :term:`CV splitter`,
          - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    train_sizes : array-like of shape (n_ticks,)
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the ``dtype`` is float, it is regarded
        as a fraction of the maximum size of the training set (that is
        determined by the selected validation method), i.e. it has to be within
        (0, 1]. Otherwise it is interpreted as absolute sizes of the training
        sets. Note that for classification the number of samples usually have
        to be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))
    """
    if axes is None:
        _, axes = plt.subplots(1, 3, figsize=(20, 5))

    axes[0].set_title(title)
    if ylim is not None:
        axes[0].set_ylim(*ylim)
    axes[0].set_xlabel("Training examples")
    axes[0].set_ylabel("Score")

    train_sizes, train_scores, test_scores, fit_times, _ = learning_curve(
        estimator,
        views,
        cv=cv,
        n_jobs=n_jobs,
        train_sizes=train_sizes,
        return_times=True,
    )
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    fit_times_mean = np.mean(fit_times, axis=1)
    fit_times_std = np.std(fit_times, axis=1)

    # Plot learning curve
    axes[0].grid()
    axes[0].fill_between(
        train_sizes,
        train_scores_mean - train_scores_std,
        train_scores_mean + train_scores_std,
        alpha=0.1,
        color="r",
    )
    axes[0].fill_between(
        train_sizes,
        test_scores_mean - test_scores_std,
        test_scores_mean + test_scores_std,
        alpha=0.1,
        color="g",
    )
    axes[0].plot(
        train_sizes, train_scores_mean, "o-", color="r", label="Training score"
    )
    axes[0].plot(
        train_sizes, test_scores_mean, "o-", color="g", label="Cross-validation score"
    )
    axes[0].legend(loc="best")

    # Plot n_samples vs fit_times
    axes[1].grid()
    axes[1].plot(train_sizes, fit_times_mean, "o-")
    axes[1].fill_between(
        train_sizes,
        fit_times_mean - fit_times_std,
        fit_times_mean + fit_times_std,
        alpha=0.1,
    )
    axes[1].set_xlabel("Training examples")
    axes[1].set_ylabel("fit_times")
    axes[1].set_title("Scalability of the model")

    # Plot fit_time vs score
    fit_time_argsort = fit_times_mean.argsort()
    fit_time_sorted = fit_times_mean[fit_time_argsort]
    test_scores_mean_sorted = test_scores_mean[fit_time_argsort]
    test_scores_std_sorted = test_scores_std[fit_time_argsort]
    axes[2].grid()
    axes[2].plot(fit_time_sorted, test_scores_mean_sorted, "o-")
    axes[2].fill_between(
        fit_time_sorted,
        test_scores_mean_sorted - test_scores_std_sorted,
        test_scores_mean_sorted + test_scores_std_sorted,
        alpha=0.1,
    )
    axes[2].set_xlabel("fit_times")
    axes[2].set_ylabel("Score")
    axes[2].set_title("Performance of the model")

    return plt


fig, axes = plt.subplots(3, 1, figsize=(10, 15))

title = "Learning Curves CCA"
# Cross validation with 50 iterations to get smoother mean test and train
# score curves, each time with 20% data randomly selected as a validation set.
cv = ShuffleSplit(n_splits=50, test_size=0.2, random_state=0)

model = CCA()
plot_learning_curve(model, title, (X, Y), axes=axes, ylim=(0.7, 1.01), cv=cv, n_jobs=4)

plt.show()



from ._elastic import ElasticCCA, SCCA_IPLS
from ._gradkcca import GradKCCA
from ._pdd import AltMaxVar
from ._pls_als import PLS_ALS
from ._pmd import SCCA_PMD
from ._scca_admm import SCCA_ADMM
from ._scca_hsic import SCCA_HSIC
from ._scca_parkhomenko import SCCA_Parkhomenko
from ._spancca import SCCA_Span
from ._swcca import SWCCA

__all__ = [
    "ElasticCCA",
    "SCCA_IPLS",
    "PLS_ALS",
    "SCCA_PMD",
    "SCCA_ADMM",
    "SCCA_Parkhomenko",
    "SCCA_Span",
    "SWCCA",
    "GradKCCA",
    "SCCA_HSIC",
    "AltMaxVar",
]



"""
Kernel CCA and Nonparametric CCA
===================================

This script demonstrates how to use kernel and nonparametric methods
"""

# %%
import numpy as np

from cca_zoo.data.simulated import LinearSimulatedData
from cca_zoo.model_selection import GridSearchCV
from cca_zoo.models import KCCA

# %%
# Data
# -----
np.random.seed(42)
n = 200
p = 100
q = 100
latent_dims = 1
cv = 3

(X, Y) = LinearSimulatedData(
    view_features=[p, q], latent_dims=latent_dims, correlation=[0.9]
).sample(n)


# %%
# Custom Kernel
def my_kernel(X, Y, param=0):
    """
    We create a custom kernel:

    """

    return np.random.normal(0, param)


kernel_custom = KCCA(
    latent_dims=latent_dims,
    kernel=[my_kernel, my_kernel],
    kernel_params=[{"param": 1}, {"param": 1}],
).fit([X, Y])

# %%
# Linear
c1 = [0.9, 0.99]
c2 = [0.9, 0.99]
param_grid = {"kernel": ["linear"], "c": [c1, c2]}
kernel_reg = GridSearchCV(
    KCCA(latent_dims=latent_dims), param_grid=param_grid, cv=cv, verbose=True
).fit([X, Y])

# %%
# Polynomial
degree1 = [2, 3]
degree2 = [2, 3]
param_grid = {"kernel": ["poly"], "degree": [degree1, degree2], "c": [c1, c2]}
kernel_poly = (
    GridSearchCV(
        KCCA(latent_dims=latent_dims), param_grid=param_grid, cv=cv, verbose=True
    )
    .fit([X, Y])
    .best_estimator_
)

# %%
# kernel cca (gaussian/rbf)
gamma1 = [1e-1, 1e-2]
gamma2 = [1e-1, 1e-2]
param_grid = {"kernel": ["rbf"], "gamma": [gamma1, gamma2], "c": [c1, c2]}
kernel_poly = (
    GridSearchCV(
        KCCA(latent_dims=latent_dims), param_grid=param_grid, cv=cv, verbose=True
    )
    .fit([X, Y])
    .best_estimator_
)


# %%
# Custom Kernel
def my_kernel(X, Y, param=0):
    """
    We create a custom kernel:

    """
    M = np.random.rand(X.shape[0], X.shape[0]) + param
    return X @ M @ M.T @ Y.T


kernel_custom = KCCA(
    latent_dims=latent_dims,
    kernel=[my_kernel, my_kernel],
    kernel_params=[{"param": 1}, {"param": 1}],
).fit([X, Y])
